try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_namelist import OMFITnamelist
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_nc import OMFITnc
from omfit_classes.omfit_mds import OMFITmdsValue, OMFITmds
from omas import ODS, omas_environment
from omfit_classes.omfit_data import Dataset, DataArray
from omfit_classes.sortedDict import OMFITdataset, SortedDict
from omfit_classes.utils_math import atomic_element
from omfit_classes.omfit_ufile import OMFITuFile
from omfit_classes.utils_plot import *

from scipy.interpolate import LinearNDInterpolator, RectBivariateSpline
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends._backend_tk import NavigationToolbar2Tk as NavigationToolbar2
import numpy as np
from scipy import integrate

__all__ = [
    'OMFITtranspNamelist',
    'OMFITtranspData',
    'OMFITtranspMultigraph',
    'OMFITplasmastate',
    'OMFITtranspOutput',
    'check_TRANSP_run',
    'wait_TRANSP_run',
    'next_available_TRANSP_runid',
    'OMFITfbm',
]

_OMFITtranspNamelist_class_attributes = {
    'collect_arrays': {'__offset__': 1},  # Collect arrays to access array data consistently within OMFIT
    'explicit_arrays': True,  # TRANSP requires arrays to have style: ARR(1)
    'compress_arrays': False,  # TRANSP does not surpport NAMELIST repetitions
    'max_array_chars': None,  # TRANSP does not support arrays on multiple lines
    'separator_arrays': ', ',
    'outsideOfNamelistIsComment': False,
    'split_arrays': True,  # TRANSP requires each namelist row not to exceed 132 characters
    'end_token': '&END',
}


class OMFITtranspNamelist(OMFITnamelist):
    r"""
    Class used to interface with TRANSP input "namelist"
    This is necessary because the TRANSP input namelist is not a default format
    (e.g. with the update_time functionality)

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, **kw):
        tmp = {}
        tmp.update(_OMFITtranspNamelist_class_attributes)
        tmp.update(kw)
        OMFITnamelist.__init__(
            self, filename, **tmp
        )  # the call to OMFITnamelist.__init__ with **kw takes care of setting up the class attributes
        for k in list(_OMFITtranspNamelist_class_attributes.keys()):
            if k in self.OMFITproperties and self.OMFITproperties[k] == _OMFITtranspNamelist_class_attributes[k]:
                self.OMFITproperties.pop(k, None)

    @dynaLoad
    def load(self):
        """
        Method used to load the content of the file specified in the .filename attribute

        :return: None
        """
        with open(self.filename, 'r') as f:
            lines = f.readlines()
        sections = SortedDict()
        sec = 'main'
        sections[sec] = []
        for line in lines:
            line = line.strip('\n')
            if re.match(r'^\s*~update_time\s*=\s*[\.0-9]', line):
                try:
                    sec = float(line.split('!')[0].split('=')[1].strip())
                    sections[sec] = []
                except Exception:
                    pass
            else:
                sections[sec].append(line)
        for sec in list(sections.keys()):
            sections[sec] = '\n'.join(sections[sec])
            tmp = NamelistFile(
                None, input_string=sections[sec], **{attr: getattr(self, attr) for attr in _OMFITtranspNamelist_class_attributes}
            )
            if sec == 'main':
                self.update(tmp)
            else:
                self[sec] = tmp

    @dynaSave
    def save(self):
        """
        Method used to save the content of the object to the file specified in the .filename attribute

        :return: None
        """
        tmp = SortedDict(sorted=True)
        for k in list(self.keys()):
            if not isinstance(k, str):
                tmp[k] = self[k]
                del self[k]
        super(OMFITnamelist, self).save()
        with open(self.filename, 'a') as f:
            for k in list(tmp.keys()):
                f.write('~update_time = %f\n' % k)
                tmp_file = NamelistFile()
                tmp_file.update(tmp[k])
                tmp_file.save(fp=f)
                self[k] = tmp_file

    @property
    def update_time(self):
        """
        This attribute returns a SortedDictionary with the update times elements in the TRANSP namelist
        """
        tmp = SortedDict(sorted=True)
        for k in sorted(self.keys()):
            if not isinstance(k, str):
                tmp[k] = self[k]
        return tmp

    def _checkSetitem(self, key, val):
        """
        Allow non-string keys
        """
        return key, val


class OMFITtranspData(SortedDict):
    """
    Class for plotting and manipulating TRANSP output MDS and CDF files
    """

    def __init__(self, transp_output, data):
        """
        Initialize data object from a OMFITmdsValue

        :param transp_output: OMFITmds object or OMFITnc object containing transp run

        :param data: Output data label

        :type data: str

        >> OMFIT['NC_d3d']=OMFITnc(filename)
        >> OMFITtranspData(OMFIT['NC_d3d'],'Q0')
        >> OMFITtranspData(OMFIT['NC_d3d'],'NE')
        >>
        >> OMFIT['MDS_d3d']=OMFITmds(server='D3D', treename='transp', shot='1551960501')
        >> OMFITtranspData(OMFIT['MDS_d3d'],'Q0')
        >> OMFITtranspData(OMFIT['MDS_d3d'],'NE')
        >>
        >> OMFIT['MDS_iter']=OMFITmds(server='transpgrid', treename='transp_iter', shot='201001676')
        >> OMFITtranspData(OMFIT['MDS_iter'],'Q0')
        >> OMFITtranspData(OMFIT['MDS_iter'],'NE')
        """

        # NetCDF
        if isinstance(transp_output, OMFITnc):
            self['DATA'] = transp_output[data]['data']
            self['DIM_OF'] = [
                transp_output[transp_output[data]['__dimensions__'][k]]['data'] for k in reversed(list(range(len(self['DATA'].shape))))
            ]
            self['UNITS_DIM_OF'] = [
                transp_output[transp_output[data]['__dimensions__'][k]]['units'].strip()
                for k in reversed(list(range(len(self['DATA'].shape))))
            ]
            self['UNITS_DIM_OF'] = [re.sub('SECONDS', 's', x) for x in self['UNITS_DIM_OF']]
            if len(self['DATA'].shape) == 2:
                self['XAXIS'] = transp_output[data]['__dimensions__'][1]
                if len(self['DIM_OF'][1].shape) == 1:
                    self['DIM_OF'] = [self['DIM_OF'][0], np.array([self['DIM_OF'][1]] * self['DIM_OF'][0].shape[1]).T]
            self['LABEL'] = transp_output[data]['long_name'].strip() + ' (%s)' % transp_output[data]['units'].strip()
            self['RPLABEL'] = transp_output[data]['long_name'].strip()
            self['UNITS'] = transp_output[data]['units'].strip()
            self['CDF'] = transp_output

        # MDS+
        elif isinstance(transp_output, OMFITmds):
            # time needs to be treated special since in MDS+ 'TIME' is stored ad TIME1D and TIME2D
            if data == 'TIME' and 'TIME' not in transp_output['OUTPUTS']['ONE_D']:
                Data = transp_output['OUTPUTS']['ONE_D']['AIMP']
                self['DATA'] = Data.dim_of(0)
                self['DIM_OF'] = [self['DATA']]
                self['UNITS_DIM_OF'] = ['s']
                self['LABEL'] = 'Time'
                self['RLABEL'] = 'Time'
                self['UNITS'] = ['s']

            # get all the data
            else:
                if data in transp_output['OUTPUTS']['ONE_D']:
                    Data = transp_output['OUTPUTS']['ONE_D'][data]
                else:
                    Data = transp_output['OUTPUTS']['TWO_D'][data]

                self['DATA'] = Data.data()

                if data == 'TIME' or data in transp_output['OUTPUTS']['ONE_D']:

                    # this is to handle different ways of storing dim_of information at GA and PPPL
                    if len(transp_output['OUTPUTS']['ONE_D']['TIME1D']):
                        time = transp_output['OUTPUTS']['ONE_D']['TIME1D'].data()
                    else:
                        time = Data.dim_of(0)
                    self['DIM_OF'] = [time]
                    self['UNITS_DIM_OF'] = ['s']

                elif data in transp_output['OUTPUTS']['TWO_D']:

                    self['XAXIS'] = Data['XAXIS'].data()[0].strip()
                    self['DIM_OF'] = [transp_output['OUTPUTS']['TWO_D'][self['XAXIS']].data()]
                    self['UNITS_DIM_OF'] = [transp_output['OUTPUTS']['TWO_D'][self['XAXIS']]['UNITS'].data()[0].strip()]

                    # this is to handle different ways of storing dim_of information at GA and PPPL
                    if len(transp_output['OUTPUTS']['ONE_D']['TIME1D']):
                        time = transp_output['OUTPUTS']['ONE_D']['TIME1D'].data()
                    else:
                        time = Data.dim_of(1)
                    self['DIM_OF'].append(np.tile(time, (Data.data().shape[1], 1)).T)
                    self['UNITS_DIM_OF'].append('s')

                # get all the metadata
                for k in ['LABEL', 'RPLABEL', 'UNITS']:
                    self[k] = Data[k].data()[0].strip()
                    self[k] = re.sub(r' \)', ')', re.sub('  +', ' ', self[k]))

            # save knowledge of the tree
            self['MDS'] = transp_output

        else:
            raise ValueError(
                'transp_output of class %s is not recognized. Must be either %s or %s' % (transp_output.__class__, OMFITmds, OMFITnc)
            )

    def plot(self, axes=None, label='RPLABEL', slice_axis=None, slice_at=[], label_show_slice=True, **kw):
        """
        Plot TRANSP data, using default metadata.

        If Data is one dimensional, it is plot using the matplotlib plot function.
        If 2D, the default is to show the data using View2d. If a slice_axis is defined, the slices
        are shown as line plots.
        Extra key words are passed to the plot or View2d function used.

        :param axes: Axes in which to make the plots.
        :type axes: Axes
        :param label: Labels the data in the plot. If 'LABEL' or 'RPLABEL' these values are taken from the Data.
        :type label: str
        :param slice_axis:  Slices 2D data along the radial (0) or time (1) axis.
        :type slice_axis: int
        :param slice_at: Slices made in slice_axis. An empty list plots all available slices.
        :type slice_at: np.ndarray

        :return: Figure
        """
        # don't modify someone's outside options
        kw = copy.copy(kw)

        # force 1D list of slices
        if slice_at is None:
            slice_at = []
        slice_at = np.array([slice_at]).ravel()

        # use metadaat for default labels
        if label in ['RPLABEL', 'LABEL']:
            if self[label].endswith('(' + self['UNITS'] + ')') or self[label].endswith('(' + self['UNITS'] + ' )'):
                label = self[label].strip()
            else:
                label = self[label] + ' (' + self['UNITS'] + ')'

        # make new plot if needed
        if axes:
            fig = axes.get_figure()
        elif 'figure' in kw:
            fig = kw['figure']
            if fig.axes:
                axes = fig.axes[0]
            else:
                axes = fig.use_subplot(111)
        else:
            fig = pyplot.gcf()
            axes = pyplot.gca()

        im_opt = {}
        if 'cmap' in kw:
            im_opt['cmap'] = kw.pop('cmap')

        # 1D data
        if len(self['DIM_OF']) == 1:
            kw.setdefault('label', label)
            axes.plot(self['DIM_OF'][0], self['DATA'], **kw)
            axes.set_xlabel('Time $[{:}]$'.format(self['UNITS_DIM_OF'][0]))

        # 2D data
        elif len(self['DIM_OF']) == 2:
            x = self['DIM_OF'][0][0]
            t = self['DIM_OF'][1][:, 0]
            y = self['DATA']

            x0label = '{:} [{:}]'.format(self['XAXIS'], self['UNITS_DIM_OF'][0]).replace('[]', '')
            x1label = '{:} [{:}]'.format('Time', self['UNITS_DIM_OF'][1]).replace('[]', '')

            # default is interactive data interface
            if slice_axis == None:
                view = View2d(
                    self['DATA'].T,
                    coords=[x, t],
                    dims=[x0label, x1label],
                    name='$[' + self['UNITS'] + ']$',
                    axes=axes,
                    imag_options=im_opt,
                    **kw,
                )
                view.colorbar.set_label(label)
                fig, axes = view.fig, view.axm
                fig.view = view

            # retain plain 2D contour plot option
            elif slice_axis < 0:
                cs = utils.utils_plot.pcolor2(x, t, self['DATA'], cmap=cmap, **kw)
                cr = utils.utils_plot.contour(x, t, y, 21, colors='k')
                cb = utils.utils_plot.colorbar(cs, label=label)
                axes.set_xlabel(x0label)
                axes.set_ylabel(x1label)

            # plot 1D slices
            else:

                # collect slices
                if len(slice_at) == 0:
                    if slice_axis == 1:
                        yy = y
                        slice_at = t
                    else:
                        yy = y.T
                        slice_at = x
                else:
                    tmp = RectBivariateSpline(x, t, y.T, kx=min(3, len(x)), ky=min(3, len(t)))
                    if slice_axis == 0:
                        yy = [tmp.ev(t * 0 + s, t) for s in slice_at]
                    elif slice_axis == 1:
                        yy = [tmp.ev(x, x * 0 + s) for s in slice_at]
                xx = [t, x][slice_axis]
                xl = [x1label, x0label][slice_axis]
                slice_label = [x1label, x0label][slice_axis - 1]

                # plot lines
                lines = []
                for yi in yy:
                    lines += axes.plot(xx, yi, **kw)
                if len(slice_at) == 1:
                    if label_show_slice:
                        label += ', {:}={:}'.format(slice_label.split()[0], slice_at[0])

                else:
                    sm = utils_plot.set_linearray(lines, slice_at)
                    cb = fig.colorbar(sm)
                    cb.set_label(slice_label)
                kw.setdefault('label', label)
                axes.lines[-1].set_label(kw['label'])
                axes.set_xlabel(xl)

                if label and label.upper() != 'tk.NONE':
                    leg = axes.legend(loc=0)
                    if leg:
                        leg.draggable(True)

        return fig


class OMFITtranspMultigraph(SortedDict):
    """
    Class for unique manipulation/plotting of TRANSP multigraph mdsValues.
    """

    def __init__(self, MDStree, data):
        """
        Initialize data object from a `OMFITmdsValue`.

        :param MDStree: OMFITmdsTree object
        :type MDStree: :class:`omfit_mds`
        :param data: Name of multigraph
        :type data: str

        """
        # get all the data
        content = list(map(lambda x: str(x).strip(), MDStree['MULTIGRAPHS'][data]['CONTENT'].data()))
        self['CONSIGN'] = np.array(MDStree['MULTIGRAPHS'][data]['CONSIGN'].data())
        self['LABEL'] = MDStree['MULTIGRAPHS'][data]['LABEL'].data()[0].strip()
        self['CONTENT'] = SortedDict()
        for k, s in zip(content, self['CONSIGN']):
            self['CONTENT'][k] = OMFITtranspData(MDStree, k)
            if np.any(self['CONSIGN'] < 0):
                self['CONTENT'][k]['DATA'] *= s
                self['CONTENT'][k]['LABEL'] = '{:}*{:}'.format(s, self['CONTENT'][k]['LABEL'])
                self['CONTENT'][k]['RPLABEL'] = '{:}*{:}'.format(s, self['CONTENT'][k]['RPLABEL'])
        self['DIM_ND'] = len(self['CONTENT'][content[-1]]['DIM_OF'])

    def plot(self, axes=None, label='LABEL', squeeze=None, **kw):
        """
        Plot each data object in the multigraph.

        :param axes: Axes in which to plot.
        :param label: String labeling the data. 'LABEL' or 'RPLABEL' are taken from TRANSP metadata.
        :param squeeze: Bool demanding all plots be made on a single figure. Default is True for 1D data.

        All other key word arguments are passed to the individual OMFITtranspData plot functions.

        """
        # defaults
        kw.setdefault('slice_at', [])
        kw.setdefault('slice_axis', None)
        kw['slice_at'] = np.array([kw['slice_at']]).ravel()  # forces 1D

        # Single lines or not
        if squeeze == None:
            if self['DIM_ND'] == 1:
                squeeze = True
            elif kw['slice_axis'] != None and len(kw['slice_at']) == 1:
                squeeze = True

        # use metadaat for default labels
        if label in ['LABEL']:
            label = self[label]

        # set up figure(s)
        if squeeze:
            if not axes:
                note, axes = pyplot.subplots()
        # make a notebook to organize/collect all the 2D plots
        else:
            master = tk.Toplevel(OMFITaux['rootGUI'])
            if OMFITaux['rootGUI'] is not None and OMFITaux['rootGUI'].globalgetvar('figsOnTop'):
                master.transient(OMFITaux['rootGUI'])
            master.wm_title(label)
            master.geometry("710x710")
            note = ttk.Notebook(master)

        # plot all content
        for k, v in list(self['CONTENT'].items()):
            # overplot the 1D quantities
            if squeeze:
                note = v.plot(axes=axes, label='LABEL', **kw)
            # pack 2D viewers into tabs in a common notebook
            else:
                # add the Tkinter GUI tab
                tab = ttk.Frame(note)
                note.add(tab, text=k)
                note.pack()
                # embed the figure
                fig = matplotlib.figure.Figure()
                canvas = FigureCanvasTkAgg(fig, master=tab)
                canvas.draw()
                fig.set_canvas(canvas)
                canvas.get_tk_widget().pack(side='top', fill='both', expand=1)
                # Add standard toolbar to plot
                toolbar = NavigationToolbar2(canvas, tab)
                toolbar.update()
                canvas._tkcanvas.pack(side='top', fill='both', expand=1)
                tab.fig = v.plot(figure=fig, **kw)

                def update_slices(tab, event=None):
                    """Keep all view slices consistent"""
                    if hasattr(tab.fig, 'view'):
                        xlim = tab.fig.view.get_vslice_args()
                        ylim = tab.fig.view.get_hslice_args()
                        for tid in note.tabs():
                            t = note.children[tid.split('.')[-1]]
                            if t != tab and hasattr(t, 'fig') and hasattr(t.fig, 'view'):
                                view = t.fig.view
                                view.hslice(*ylim)
                                view.vslice(*xlim)

                tab.update_slices = utils.types.MethodType(update_slices, tab)
                tab.bind('<FocusOut>', tab.update_slices)

        # extra features
        if squeeze:
            leg = axes.legend()
            leg.draggable(True)
        else:
            # add summary tab to the Tkinter GUI
            if hasattr(tab.fig, 'view'):
                tab = ttk.Frame(note)
                note.add(tab, text='Summary')
                note.pack()
                # embed the figure
                fig = matplotlib.figure.Figure()
                canvas = FigureCanvasTkAgg(fig, master=tab)
                canvas.draw()
                fig.set_canvas(canvas)
                canvas.get_tk_widget().pack(side='top', fill='both', expand=1)
                # Add standard toolbar to plot
                toolbar = NavigationToolbar2(canvas, tab)
                toolbar.update()
                canvas._tkcanvas.pack(side='top', fill='both', expand=1)
                axx = fig.add_subplot(211)
                axy = fig.add_subplot(212)
                tab.fig = fig

                def update_lines(tab, event=None):
                    axx, axy = tab.fig.axes[:2]
                    for a in [axx, axy]:
                        a.lines = []
                        a.collections = []
                        a.set_prop_cycle('color', pyplot.rcParams['axes.prop_cycle'].by_key()['color'])
                    for tid in note.tabs():
                        k = note.tab(tid, 'text')
                        t = note.children[tid.split('.')[-1]]
                        if not hasattr(t, 'fig') or not hasattr(t.fig, 'view'):
                            continue
                        fig = t.fig
                        v = fig.view
                        # re-produce fills
                        (l,) = axx.plot(v._hslice_xdata, v._hslice_mean, label=k, **v.plot_options)
                        axx.fill_between(
                            v._hslice_xdata, v._hslice_mean - v._hslice_std, v._hslice_mean + v._hslice_std, color=l.get_color(), alpha=0.3
                        )
                        (l,) = axy.plot(v._vslice_xdata, v._vslice_mean, label=k, **v.plot_options)
                        axy.fill_between(
                            v._vslice_xdata, v._vslice_mean - v._vslice_std, v._vslice_mean + v._vslice_std, color=l.get_color(), alpha=0.3
                        )
                    # refresh labels
                    # axx.texts = [axx.text(0.02,0.95,v.axx.texts[0].get_text(),verticalalignment='top',
                    #                      transform=axx.transAxes)]
                    # axy.texts = [axy.text(0.02,0.95,v.axy.texts[0].get_text(),verticalalignment='top',
                    #                      transform=axy.transAxes)]
                    axx.set_xlabel(fig.view.axm.get_xlabel())
                    axx.set_ylabel(fig.view.axx.get_ylabel())
                    axy.set_ylabel(fig.view.axy.get_xlabel())
                    axy.set_xlabel(fig.view.axm.get_ylabel())
                    axx.autoscale()
                    axy.autoscale()
                    lx = axx.legend()
                    ly = axy.legend()
                    tab.fig.canvas.draw()
                    return

                tab.update_lines = utils.types.MethodType(update_lines, tab)
                tab.bind('<FocusIn>', tab.update_lines)

            # resize notebook figures with main window
            def set_fig_size(event=None):
                master.unbind('<Configure>')
                master.update_idletasks()
                note.configure(width=master.winfo_width())
                note.configure(height=master.winfo_height())
                master.bind('<Configure>', set_fig_size)

            set_fig_size()
            master.bind('<Configure>', set_fig_size)

            # allow keys to change tabs
            note.enable_traversal()

        return note

    plot.__doc__ += OMFITtranspData.plot.__doc__


def check_TRANSP_run(runid, project=None, tunnel=None):
    """
    Function that checks the status of a TRANSP run as reported
    by the TRANSP MONITOR GRID website: https://w3.pppl.gov/transp/transpgrid_monitor

    :param runid: runid to be checked

    :param project: project (ie. tokamak) of the runid (optional)

    :param tunnel: use SOCKS via specified tunnel

    :return: * None if no matching runid/project is found
             * tuple with 1) True/None/False if run went ok, run is waiting, run failed
                     and  2) dictionary with parsed HTML information
    """

    socks = {}
    if tunnel is not None:
        tmp = setup_socks(tunnel)
        socks['http'] = 'socks5://localhost:' + tmp[2]
        socks['https'] = 'socks5://localhost:' + tmp[2]
    url = 'https://w3.pppl.gov/cgi-bin/transpgrid_monitor.frames'
    data = {'runid': runid}
    if project is not None:
        data.update({'project': project})

    import requests
    from urllib3.exceptions import InsecureRequestWarning

    with warnings.catch_warnings(record=False) as w:
        warnings.filterwarnings("ignore", category=InsecureRequestWarning)
        response = requests.get(url, data, proxies=socks, verify=False)
    response.raise_for_status()
    the_page = response.text

    items = ['project', 'owner', 'details', 'status', 'remarks']

    parsed = OrderedDict()
    for line in the_page.split('\n'):
        if np.all([k in line for k in items]):
            line = re.sub('&nbsp', '', line)
            line = re.sub('''<(?:"[^"]*"['"]*|'[^']*'['"]*|[^'">])+>''', '\n', line)
            line = [_f for _f in line.split('\n') if _f]
            for entry, value in zip(line[::2], line[1::2]):
                parsed.setdefault(entry, '')
                parsed[entry] += value

    wait = ['submitted', 'active', 'HALT_RQST', 'mdswrite']
    ok = ['success', 'Fetched']
    bad = ['stopped', 'canceled', 'HALT_FAILED', 'double', 'nopriv', 'suspended']

    if np.any([k in ok for k in list(parsed.keys())]):
        return True, parsed

    elif np.any([k in wait for k in list(parsed.keys())]):
        return None, parsed

    elif np.any([k in bad for k in list(parsed.keys())]):
        return False, parsed


def wait_TRANSP_run(runid, project=None, t_check=5, verbose=True, tunnel=None):
    """
    Function that waits for a TRANSP run to end as reported
    by the TRANSP MONITOR GRID website: https://w3.pppl.gov/transp/transpgrid_monitor

    :param runid: runid to be checked

    :param project: project (ie. tokamak) of the runid (optional)

    :param t_check: how often to check (seconds)

    :param verbose: print to screen

    :param tunnel: use SOCKS via specified tunnel

    :return: * None if no matching runid/project is found
             * tuple with 1) True/False if run went ok or run failed
                     and  2) dictionary with parsed HTML information
    """

    aborted = tk.BooleanVar()
    aborted.set(False)
    if OMFITaux['rootGUI'] is not None:

        def onAbort():
            aborted.set(True)

        try:
            if OMFITaux['pythonRunWindows'][-1] is None:
                raise RuntimeError('--')
            top = ttk.Frame(OMFITaux['pythonRunWindows'][-1], borderwidth=2, relief=tk.GROOVE)
            top.pack(side=tk.TOP, expand=tk.NO, fill=tk.BOTH, padx=5, pady=5)
        except Exception:
            top = tk.Toplevel(OMFITx._topGUI(OMFITaux['rootGUI']))
            top.transient(OMFITaux['rootGUI'])

        top.update_idletasks()
        ttk.Label(top, text='TRANSP wait ' + str(runid), wraplength=top.winfo_width()).pack(side=tk.TOP, expand=tk.NO, fill=tk.X)
        p = ttk.Progressbar(top, orient=tk.HORIZONTAL, mode='indeterminate')
        p.start()
        p.pack(padx=5, pady=5, expand=tk.NO, fill=tk.X)
        frm = ttk.Frame(top)
        ttk.Button(frm, text="Abort", command=onAbort).pack(side=tk.LEFT, expand=tk.NO, fill=tk.X)
        frm.pack(side=tk.TOP)
        top.update_idletasks()

    status = None
    old_print = None
    t = -1
    dt = 0.1
    n = 0
    while not aborted.get():
        if t < 0:
            t = t_check
            out = check_TRANSP_run(runid, project, tunnel=tunnel)

            # if the runid does not appear in the list
            # wait 5x the t_check before giving up
            if out is None:
                if n == 5:
                    if OMFITaux['rootGUI'] is not None:
                        top.destroy()
                    return None
                n += 1

            # if the runid is listed print only new updates
            else:
                n = 0
                if old_print != repr(out[1]):
                    if verbose:
                        print('-' * 5 + ' ' + str(runid) + ' ' + '-' * 5 + ' ' + now())
                        for k in out[1]:
                            printi('%s : %s' % (k, out[1][k]))
                elif out[0] is not None:
                    break
                old_print = repr(out[1])

        sleep(dt)

        t = t - dt

    if OMFITaux['rootGUI'] is not None:
        top.destroy()

    if aborted.get():
        raise OMFITexception('-- Aborted by user waiting for TRANSP run %s --' % str(runid))

    return out


def next_available_TRANSP_runid(runid, project, increment=1, skiplist=[], server=None):
    """
    Function which checks what MDS+ tree entries are available

    :param runid: runid to start checking from (included)

    :param project: project [e.g. D3D, ITER, NSTX, ...] used to set the MDS+ TRANSP tree and server to be queried

    :param increment: positive / negative

    :param skiplist: list of runids to skip

    :param server: MDS+ TRANSP server to be queried [e.g. 'atlas.gat.com', or 'transpgrid.pppl.gov']

    :return: tuple with next available runid (as integer with format shotXXYY) and augmented skiplist
    """

    # handle letter in runid
    try:
        int(runid)
    except Exception:
        shot = runid[:-3]
        letter = runid[-3:-2].upper()
        index = runid[-2:]
        runid = str(str(shot) + str(ord(letter) - ord('A') + 1) + index)
    runid = str(runid)
    runid_base = str(runid)[:-2]

    # wrap around counting up/down
    if increment > 0:
        lst = np.mod(int(runid[-2:]) - 1 + np.arange(0, 99), 99) + 1
    else:
        lst = np.mod(int(runid[-2:]) - 1 - np.arange(0, 99), 99) + 1

    # handle MDS+ server/treename
    if server == None:
        if project in ['D3D', 'DIII-D']:
            server = 'atlas.gat.com'
            project = 'D3D'
        else:
            server = 'transpgrid.pppl.gov'
    if server == 'atlas.gat.com':
        tree = 'transp'
    else:
        tree = 'transp_' + project.lower()

    # look in search for first available spot
    list_busy = copy.copy(skiplist)
    import MDSplus

    for k in lst:
        runid = int('%s%02d' % (runid_base, k))
        if k in list_busy:
            printi('skipping %d' % runid)
            continue
        printi('testing  %d' % runid)
        try:
            tmp = OMFITmdsValue(server=server, treename=tree, shot=runid, TDI='\\' + tree.upper() + '::TOP.OUTPUTS.ONE_D.AIMP').data()
            if tmp is None or isinstance(tmp, np.ndarray):
                printw('MDSplus tree exists')
            list_busy.append(k)
        except Exception as _excp:
            printi('Found available MDSplus tree')
            return runid, list_busy

    raise OMFITexception('No available runid for %s' % runid_base)


class OMFITplasmastate(OMFITnc):
    """
    Class for handling TRANSP netcdf statefile (not to be confused with the time-dependent TRANSP output CDF file)
    """

    sources = {
        'pe_trans': 'Total power to electrons',
        'pi_trans': 'Total power to ions',
        'qie': 'Collisional exchange from ions to electrons',
        'pbe': 'Beam power to electrons',
        'pbi': 'Beam power to ions',
        'pbth': 'Thermalization of beam power to ions',
        'peech': 'ECH power to electrons',
        'pohme': 'Ohmic heating power to electrons',
        'pmine': 'Electron heating power by minority ions',
        'pmini': 'Ion heating power by minority ions',
        'pminth': 'Thermalization of ion heating power by minority ions',
        'picth': 'Direct ion heating power by ICRF',
        'pfuse': 'Fusion alpha power transferred to electrons',
        'pfusi': 'Fusion alpha power transferred to thermal ions',
        'pfusth': 'Thermalization of fusion alpha power transferred to thermal ions',
        'prad_cy': 'Radiated power: synchrotron',
        'prad_br': 'Radiated power: bremsstrahlung',
        'prad_li': 'Radiated power: line',
        'tq_trans': 'Angular momentum source torque',
        'sn_trans': 'Particle source',
    }

    def calcQ(self):
        """
        :return: fusion gain
        """
        auxi = 0
        for k in ['pbi', 'picth', 'pmini']:
            auxi += np.sum(self[k]['data'])

        auxe = 0
        for k in ['pbe', 'peech', 'pmine', 'pohme']:
            auxe += np.sum(self[k]['data'])

        aux = auxe + auxi

        fuse = np.sum(self['pfuse']['data'])
        fusi = np.sum(self['pfusi']['data'])
        fus_alpha = fuse + fusi
        fus_neut = (fuse + fusi) * 4

        return (fus_neut + fus_alpha) / aux

    def summary_sources(self):
        """
        Print summary of integrated sources
        """
        for item in sorted(list(self.sources.keys())):
            if item.startswith('p'):
                print('%s:% 4d MW      [%s]' % (item.ljust(10), np.sum(self[item]['data']) / 1e6, self.sources[item]))
            if item.startswith('s'):
                print('%s:% 4d 1E20/s  [%s]' % (item.ljust(10), np.sum(self[item]['data']) / 1e20, self.sources[item]))
            if item.startswith('t'):
                print('%s:% 4d N*m^2   [%s]' % (item.ljust(10), int(np.sum(self[item]['data'])), self.sources[item]))

    def to_omas(self, ods=None, time_index=0, update=['core_profiles', 'equilibrium', 'core_sources']):
        """
        translate TRANSP plasmastate file (output of TRXPL and plasmastate of SWIM) to OMAS data structure

        :param ods: input ods to which data is added

        :param time_index: time index to which data is added

        :update: list of IDS to update from statefile

        :return: ODS
        """
        from omas import ODS, omas_environment

        cocosio = 2  # need to confirm that indeed TRANSP plasmastate is in COCOS 2

        if ods is None:
            ods = ODS()
            # set the shot number
            ods['dataset_description.data_entry.pulse'] = self['shot_number']['data']

        # ------------------
        # Equilibrium
        # ------------------
        if 'equilibrium' in update:
            # set time array
            ods.set_time_array('equilibrium.time', time_index, self['t1']['data'])  # t0,t1,tfinal,tinit?

            # shortcut
            eq = ods['equilibrium.time_slice'][time_index]

            # define coordinate of input/output quantities
            coordsio = {'equilibrium.time_slice.%d.profiles_1d.psi' % time_index: self['psipol']['data']}

            # assign data
            with omas_environment(ods, cocosio=cocosio, coordsio=coordsio):
                eq['global_quantities.magnetic_axis.b_field_tor'] = self['B_axis_vac']['data']

                ods.set_time_array('equilibrium.vacuum_toroidal_field.b0', time_index, self['B_axis_vac']['data'])
                ods['equilibrium.vacuum_toroidal_field.r0'] = self['R_axis']['data']

                eq['global_quantities.magnetic_axis.r'] = self['R_axis']['data']
                eq['global_quantities.magnetic_axis.z'] = self['Z_axis']['data']
                eq['global_quantities.psi_axis'] = self['psipol']['data'][0]
                eq['global_quantities.psi_boundary'] = self['psipol']['data'][-1]
                eq['global_quantities.ip'] = self['curt']['data'][-1]

                eq['profiles_1d.rho_tor_norm'] = self['rho']['data']

                eq['profiles_1d.f'] = self['g_eq']['data']
                eq['profiles_1d.pressure'] = self['P_eq']['data']
                eq['profiles_1d.f_df_dpsi'] = deriv(self['psipol']['data'], self['g_eq']['data']) * self['g_eq']['data']
                eq['profiles_1d.dpressure_dpsi'] = deriv(self['psipol']['data'], self['P_eq']['data'])
                eq['profiles_1d.q'] = self['q_eq']['data']
                eq['profiles_1d.rho_tor_norm'] = self['rho']['data']

                eq['profiles_2d.0.grid_type.index'] = 1
                eq['profiles_2d.0.grid.dim1'] = self['R_grid']['data']
                eq['profiles_2d.0.grid.dim2'] = self['Z_grid']['data']
                eq['profiles_2d.0.psi'] = self['PsiRZ']['data'].T

                eq['boundary.outline.r'] = self['R_geo']['data'][:, -1]
                eq['boundary.outline.z'] = self['Z_geo']['data'][:, -1]

                eq['profiles_1d.elongation'] = self['elong']['data']
                eq['profiles_1d.triangularity_upper'] = self['triangU']['data']
                eq['profiles_1d.triangularity_lower'] = self['triangL']['data']

            # ============WALL
            ods['wall.description_2d.0.limiter.type.name'] = 'first_wall'
            ods['wall.description_2d.0.limiter.type.index'] = 0
            ods['wall.description_2d.0.limiter.type.description'] = 'first wall'
            ods['wall.description_2d.0.limiter.unit.0.outline.r'] = self['rlim']['data']
            ods['wall.description_2d.0.limiter.unit.0.outline.z'] = self['zlim']['data']

        # ------------------
        # Core profiles
        # ------------------
        if 'core_profiles' in update:
            # set time array
            ods.set_time_array('core_profiles.time', time_index, self['t1']['data'])  # t0,t1,tfinal,tinit?

            # shortcut
            prof1d = ods['core_profiles.profiles_1d.%d' % time_index]

            # define coordinate of input/output quantities
            coordsio = {
                'core_profiles.profiles_1d.%d.grid.rho_tor_norm' % time_index: self['rho']['data'][:-1] + np.diff(self['rho']['data']) / 2.0
            }

            # assign data
            with omas_environment(ods, coordsio=coordsio, cocosio=cocosio):
                ods.set_time_array('core_profiles.vacuum_toroidal_field.b0', time_index, self['B_axis_vac']['data'])
                ii = -1
                fi = -1
                bi = -1
                mi = -1
                for ks, species in enumerate(self['ALL_name']['data']):
                    species = species.strip()
                    if species == 'e':
                        prof1d['electrons.density_thermal'] = self['ns']['data'][ks, :]
                        prof1d['electrons.temperature'] = self['Ts']['data'][ks, :] * 1e3
                    else:
                        ii += 1
                        if '_fusn' in species:
                            fi += 1
                            prof1d['ion'][ii]['label'] = species.replace('_fusn', '')
                            prof1d['ion'][ii]['density_fast'] = self['nfusi']['data'][fi, :]
                            prof1d['ion'][ii]['pressure_fast_perpendicular'] = self['eperp_fusi']['data'][fi, :]
                            prof1d['ion'][ii]['pressure_fast_parallel'] = self['epll_fusi']['data'][fi, :]
                        elif '_beam' in species:
                            bi += 1
                            prof1d['ion'][ii]['label'] = species.replace('_beam', '')
                            prof1d['ion'][ii]['density_fast'] = self['nbeami']['data'][bi, :]
                            prof1d['ion'][ii]['pressure_fast_perpendicular'] = self['eperp_beami']['data'][bi, :]
                            prof1d['ion'][ii]['pressure_fast_parallel'] = self['epll_beami']['data'][bi, :]
                        elif '_mini' in species:
                            mi += 1
                            prof1d['ion'][ii]['label'] = species.replace('_mini', '')
                            prof1d['ion'][ii]['density_fast'] = self['nmini']['data'][bi, :]
                            prof1d['ion'][ii]['pressure_fast_perpendicular'] = self['eperp_mini']['data'][bi, :]
                            prof1d['ion'][ii]['pressure_fast_parallel'] = self['epll_mini']['data'][bi, :]
                        else:
                            prof1d['ion'][ii]['label'] = species
                            prof1d['ion'][ii]['density_thermal'] = self['ns']['data'][ks, :]
                            prof1d['ion'][ii]['temperature'] = self['Ts']['data'][ks, :]
                        prof1d['ion'][ii]['element'][0]['z_n'] = list(atomic_element(symbol=prof1d['ion'][ii]['label']).values())[0]['Z']
                        prof1d['ion'][ii]['element'][0]['a'] = list(atomic_element(symbol=prof1d['ion'][ii]['label']).values())[0]['A']
                        prof1d['ion'][ii]['multiple_states_flag'] = 0

            # calculate derived quantites
            ods.physics_core_profiles_pressures()
            ods.physics_core_profiles_zeff()

        # ------------------
        # Core sources
        # ------------------
        if 'core_sources' in update:

            # set time array
            ods.set_time_array('core_sources.time', time_index, self['t1']['data'])  # t0,t1,tfinal,tinit?

            def set_source(si, source_name, id_index, source_value, destination):
                # define coordinates for all sources
                coordsio = {}
                coordsio['core_profiles.profiles_1d.%d.grid.rho_tor_norm' % (time_index)] = (
                    self['rho']['data'][:-1] + np.diff(self['rho']['data']) / 2.0
                )
                coordsio['core_sources.source.%d.profiles_1d.%d.grid.rho_tor_norm' % (si, time_index)] = (
                    self['rho']['data'][:-1] + np.diff(self['rho']['data']) / 2.0
                )
                with omas_environment(ods, coordsio=coordsio, cocosio=cocosio):
                    ods['core_sources.source.%d.identifier.name' % si] = source_name
                    ods['core_sources.source.%d.identifier.index' % si] = id_index
                    source1d = ods['core_sources.source.%d.profiles_1d.%d' % (si, time_index)]
                    source1d['grid.volume'] = vol
                    source1d[destination] = source_value

            # fluxes are defined between nodes in the grid
            # volume is stored at the nodes of the grid
            # place volume info on same grid as fluxes
            vol = interp1d(self['rho']['data'], self['vol']['data'])(self['rho']['data'][:-1] + np.diff(self['rho']['data']) / 2.0)

            # assign data

            prof1d['grid.volume'] = vol

            volumetric_electron_heating_terms = {'pohme': 7, 'pbe': 2, 'peech': 3, 'pmine': 5, 'pfuse': 6}
            volumetric_ion_heating_terms = {'pbi': 2, 'picth': 5, 'pmini': 5, 'pfusi': 6}
            volumetric_electron_particles_terms = {'sbsce': 2}
            volumetric_momentum_terms = {'tq_trans': 1}

            si = -1
            # electron energy
            for source, id_index in volumetric_electron_heating_terms.items():
                if source in self:
                    si += 1
                    set_source(si, source, id_index, self[source]['data'], 'electrons.energy')

            # ion energy
            for source, id_index in volumetric_ion_heating_terms.items():
                if source in self:
                    si += 1
                    set_source(si, source, id_index, self[source]['data'], 'total_ion_energy')

            # particle ???
            for source, id_index in volumetric_electron_particles_terms.items():
                if source in self:
                    si += 1
                    set_source(si, source, id_index, self[source]['data'][0, :], 'electrons.particles')

            # momentum ???
            for source, id_index in volumetric_momentum_terms.items():
                if source in self:
                    si += 1
                    set_source(si, source, id_index, self[source]['data'], 'momentum_tor')

        return ods


class transp_out_dynamic_quantity(object):
    def __init__(self, parent, name):
        self.parent = parent
        self.name = name

    def __call__(self):
        return self.parent[self.name]

    def __tree_repr__(self):
        return self.name, []


class OMFITtranspOutput(OMFITdataset):
    """
    Class for dynamic serving of TRANSP output data from MDS or CDF
    """

    def __init__(self, transp_out):
        """
        :param transp_out: OMFITnc file, OMFITmds TRANSP tree, or string path to NetCDF file
        """
        if isinstance(transp_out, str):
            transp_out = OMFITnc(transp_out)
        if isinstance(transp_out, OMFITmds):
            self.type = 'MDS'
        else:
            self.type = 'CDF'
        self.transp_out = transp_out
        self.dynaLoad = True
        self._dynamic_keys = []
        OMFITdataset.__init__(self)

    @dynaLoad
    def load(self):
        if self.type == 'MDS':
            items = self.transp_out['TRANSP_OUT'].keys()
        else:
            items = self.transp_out.keys()
        items = [item for item in items if not item.startswith('_')]
        for item in items:
            self._dynamic_keys.append(item)
        if 'TIME' not in self._dynamic_keys:
            self._dynamic_keys.insert(0, 'TIME')

    def __tree_repr__(self):
        return self.type, []

    @dynaLoad
    def __tree_keys__(self):
        return np.unique(list(self.variables.keys()) + self._dynamic_keys).tolist()

    @dynaLoad
    def __getitem__(self, key):
        # return items that exist
        if key in self:
            return OMFITdataset.__getitem__(self, key)

        # return virtual items if requested so
        elif OMFITaux['virtualKeys'] and key in self._dynamic_keys:
            return transp_out_dynamic_quantity(self, key)

        # dynamically evaluate quantity
        elif key in self._dynamic_keys:
            transp_data = OMFITtranspData(self.transp_out, key)
            # 0D data
            if not len(transp_data['DIM_OF']):
                tmp = DataArray(transp_data['DATA'], coords={}, dims=())
            # 1D data
            elif len(transp_data['DIM_OF']) == 1:
                tmp = DataArray(transp_data['DATA'], coords={'TIME': transp_data['DIM_OF'][0]}, dims=('TIME',))
            # 2D data
            elif len(transp_data['DIM_OF']) == 2:
                x = transp_data['DIM_OF'][0][0]
                t = transp_data['DIM_OF'][1][:, 0]
                y = transp_data['DATA']
                tmp = DataArray(transp_data['DATA'], coords={'TIME': t, 'X': x}, dims=('TIME', 'X'))

            self[key] = tmp
        return OMFITdataset.__getitem__(self, key)

    def to_omas(self):
        # this may be of use: https://github.com/transp/transp-imas-translator/blob/master/transp2imas/transp2imas.f90
        ods = ODS()

        with omas_environment(ods, cocosio=2):  # what COCOS is TRANSP using?
            ion_species = ['D']  # what is a good way to figure out ion species?
            times = self['TIME']

            for time_index, time in enumerate(times):
                # ion-profiles
                for ion_index, ion_species in enumerate(ion_species):
                    ods['core_profiles.profiles_1d.{time_index}.ion.{ion_index}.temperature'.format(**locals())] = self['TI'][time_index, :]

                # electron-profiles
                ods['core_profiles.profiles_1d.{time_index}.electrons.density_thermal'.format(**locals())] = self['NE'][time_index, :] * 1e6
                ods['core_profiles.profiles_1d.{time_index}.electrons.temperature'.format(**locals())] = self['TE'][time_index, :]

                # equilibrium
                ods['equilibrium.time_slice.{time_index}.global_quantities.ip'.format(**locals())] = self['PCURC'][time_index]

                # wall: for the the time being, we only save limiter info at first time-slice
                #       hopefully other codes can agree this is reasonable for static walls
                if time_index == 0:
                    ods['wall.description_2d.{time_index}.limiter.type.name'.format(**locals())] = 'first_wall'
                    ods['wall.description_2d.{time_index}.limiter.type.index'.format(**locals())] = 0
                    ods['wall.description_2d.{time_index}.limiter.type.description'.format(**locals())] = 'first wall'
                    ods['wall.description_2d.{time_index}.limiter.unit.0.outline.r'.format(**locals())] = self['RLIM'][time_index, :]
                    ods['wall.description_2d.{time_index}.limiter.unit.0.outline.z'.format(**locals())] = self['YLIM'][time_index, :]

            # set the time
            ods['core_profiles.time'] = times
            ods['equilibrium.time'] = times
            ods['wall.time'] = times[:1]

            # set homogeneous time and other things
            ods.satisfy_imas_requirements()
        return ods


class OMFITfbm(OMFITnc):
    """
    Class for handling NUBEAM FBM distribution function files
    """

    def plot(self, rf=None, zf=None, cartesian=True, fig=None):
        """
        Plot distribution function

        :param rf: radial location where to show data in velocity space

        :param zf: vertical location where to show data in velocity space

        :return: figure handle
        """
        f = self['F_D_NBI']['data']
        energy = self['E_D_NBI']['data']
        pitch = self['A_D_NBI']['data']
        r2d = self['R2D']['data']
        z2d = self['Z2D']['data']
        rho = self['X2D']['data']
        theta = self['TH2D']['data']

        # Find index of our desired (R,Z)
        if rf is None or zf is None:
            idx = 0
        else:
            idx = np.argmin(np.sqrt((r2d - rf) ** 2 + np.abs(z2d - zf) ** 2))

        # Get beam ion density (#/cm^3)
        de = energy[1] - energy[0]
        dp = pitch[1] - pitch[0]
        bdens = np.sum(f, axis=(1, 2)) * 0.5 * de * dp

        # Form regular grid
        rr = np.linspace(np.min(r2d), np.max(r2d), 51)
        zr = np.linspace(np.min(z2d), np.max(z2d), 51)
        bdensr = scipy.interpolate.griddata((r2d, z2d), bdens, (rr[None, :], zr[:, None]), method='cubic')

        if fig is None:
            fig = pyplot.gcf()
        fig.suptitle(
            '{}'.format(self['TOKAMAK']['data'])
            + ' {}'.format(self['TRANSP_RUNID']['data'])
            + ' t={0:0.2f}s'.format(float(self['TIME']['data']))
        )

        ax = pyplot.subplot(1, 2, 1)
        con = ax.contourf(rr, zr, bdensr, 11)
        ax.contour(rr, zr, bdensr, 11, colors='black')
        ax.plot(r2d[idx], z2d[idx], marker='x', color='red')
        ax.set_aspect('equal')
        ax.set_title(label='Density $(cm^{-3})$')
        ax.set_xlabel('R (cm)')
        ax.set_ylabel('Z (cm)')
        fig.subplots_adjust(left=0.2)
        cax = fig.add_axes([0.05, 0.1, 0.03, 0.8])
        fig.colorbar(con, cax=cax)

        ax = pyplot.subplot(2, 2, 2)

        if not cartesian:
            t = np.arccos(np.linspace(-1, 1, len(pitch)))
            e = energy * 1e-3
            E, T = np.meshgrid(e, t)
            x = np.cos(T) * E
            y = np.sin(T) * E
            ax.contourf(x, y, f[idx, :, :], 11)
            ax.set_aspect('equal')
            ax.set_xlabel('Energy$_\\parallel$ (keV)')
            ax.set_ylabel('Energy$_\\perp$ (keV)')
        else:
            ax.contourf(energy * 1e-3, pitch, f[idx, :, :], 11)
            ax.set_ylim(-1, 1)
            ax.set_xlabel('Energy (keV)')
            ax.set_ylabel('Pitch $V_{||}/V$')
        ax.set_title('$f_b(E,\\xi)$' + ' R={0:0.1f}, Z={1:0.1f}'.format(r2d[idx], z2d[idx]))

        ax = pyplot.subplot(2, 2, 4)
        ax.scatter(rho, bdens * 1e-13, c=theta, alpha=0.25, label='$n_b(\\theta) [10^{19}m^{-3}]$')
        ax.set_xlim([0.0, 1.0])
        ax.set_ylim([0.0, np.max(bdens) * 1e-13])
        ax.set_xlabel('$\\rho$')
        ax.legend()

        return fig

    def plot_energy_space(self, rmin, rmax, zmin, zmax, emin=0 * 1e3, emax=1000 * 1e3, cartesian=True, ax=None):
        """
        Average distribution function over a specified R,Z and plot energy versus pitch

        :param rmin: minimum R to average over

        :param rmax: maximum R to average over

        :param zmin: minimum Z to average over

        :param zmax: maximum Z to average over

        :param emin: minimum energy to average over

        :param emax: maximum energy to average over

        :param cartesian: plot in energy/pitch space or E_\\parallel and E_\\perp

        :param ax: axes
        """

        f = self['F_D_NBI']['data']
        R = self['R2D']['data']
        Z = self['Z2D']['data']
        energy = self['E_D_NBI']['data']
        pitch = self['A_D_NBI']['data']

        emini = np.argmin(abs(energy - emin))
        emaxi = np.argmin(abs(energy - emax))

        total_f = np.zeros((f.shape[1], f.shape[2]))

        k = 0
        for i, r in enumerate(R):
            if r <= rmax and r >= rmin:
                if Z[i] <= zmax and Z[i] >= zmin:
                    total_f[:, :] += f[i, :, :]
                    k += 1

        levels = [
            np.amin(total_f[:, emini:emaxi] / k),
            np.amax(total_f[:, emini:emaxi] / k) / 40,
            np.amax(total_f[:, emini:emaxi] / k) / 20,
            np.amax(total_f[:, emini:emaxi] / k) / 10,
            np.amax(total_f[:, emini:emaxi] / k) / 4,
            np.amax(total_f[:, emini:emaxi] / k) / 3,
            np.amax(total_f[:, emini:emaxi] / k) / 2,
            np.amax(total_f[:, emini:emaxi] / k) / 1.5,
            np.amax(total_f[:, emini:emaxi] / k) / 1.25,
            np.amax(total_f[:, emini:emaxi] / k) / 1.2,
            np.amax(total_f[:, emini:emaxi] / k) / 1.1,
            np.amax(total_f[:, emini:emaxi] / k) / 1.05,
            np.amax(total_f[:, emini:emaxi] / k),
        ]

        if ax is None:
            ax = pyplot.gca()

        if not cartesian:
            t = np.arccos(np.linspace(-1, 1, len(pitch)))
            e = energy[emini:emaxi] * 1e-3
            E, T = np.meshgrid(e, t)
            x = np.cos(T) * E
            y = np.sin(T) * E
            ax.contourf(x, y, total_f[:, emini:emaxi] / k, levels=levels)
            ax.contour(x, y, total_f[:, emini:emaxi] / k, levels=levels)
            ax.set_aspect('equal')
            ax.set_xlabel('Energy$_\\parallel$ (keV)')
            ax.set_ylabel('Energy$_\\perp$ (keV)')
        else:
            ax.contourf(energy[emini:emaxi] * 1e-3, pitch, total_f[:, emini:emaxi] / k, levels=levels)
            ax.set_ylim(-1, 1)
            ax.set_xlim(10, 100)
            ax.set_xlabel('Energy (keV)')
            ax.set_ylabel('Pitch $V_\\parallel/V$')
        ax.set_title(
            'D3D $f_b(E,\\xi)$ '
            + ' {0} Rmin={1:0.1f}, Rmax={2:0.1f}, \nZmin={3:0.1f}, Zmax={4:0.1f}, average over {5} points'.format(
                self['TRANSP_RUNID']['data'], rmin, rmax, zmin, zmax, k
            )
        )
        return ax

    def to_omas(self, ods=None, time_index=0):
        """
        Save NUBEAM distribution function to OMAS

        :param ods: input ods to which data is added

        :param time_index: time index to which data is added

        :return: updated ODS
        """
        if ods is None:
            ods = ODS()

        # shortcut
        ggd = ods[f'distributions.distribution[{time_index}].ggd[0]']

        # assign grid spaces for different spatial coordinates
        for s, (item, name, coordinates_type, norm) in enumerate(
            [
                ('R2D', 'r', 1, 0.01),  # r [cm-->m]
                ('Z2D', 'z', 2, 0.01),  # z [cm-->m]
                ('X2D', 'rho', 12, 1.0),  # rho [-]
                ('TH2D', 'theta', 22, 1.0),  # theta [rad]
            ]
        ):
            # standard or fourier data
            ggd[f'grid.space[{s}].geometry_type.index'] = 0  # 0: standard, 1:Fourier
            ggd[f'grid.space[{s}].geometry_type.name'] = name + '_pitch_energy'
            ggd[f'grid.space[{s}].coordinates_type'] = [coordinates_type, 403, 301]
            ggd[f'grid.space[{s}].objects_per_dimension[0].object[0].geometry'] = self[item]['data'] * norm
            ggd[f'grid.space[{s}].objects_per_dimension[0].object[1].geometry'] = self['A_D_NBI']['data']  # pitch v_parallel/v [-]
            ggd[f'grid.space[{s}].objects_per_dimension[0].object[2].geometry'] = self['E_D_NBI']['data'] * 1e3  # energy [kev --> ev]

        # assign the distribution function data itself
        ggd[f'expansion[0].grid_subset[0].values'] = self['F_D_NBI']['data'].flatten()

        return ods


################################################# Basic Data Manipulation in TRANSP


def vint(d, dvol=None):
    """
    Volume integrate a TRANSP OMFITmdsValue object.
    Currently only available for objects from the TRANSP tree.

    :param d: OMFITtranspData object from the mds+ TRANSP OUTPUTS.TWO_D tree.
    :type d: OMFITtranspData

    :param dvol: OMFITtranspData object 'dvol' from the mds TRANSP tree.
        If None, will be taken from Data's MDStree.
    :type dvol: OMFITtranspData or None

    **Example:**
    Assuming the root is an OMFIT TRANSP module with a loaded run.
    >> mvisc = OMFITtranspData(root['OUTPUTS']['TRANSP_OUTPUT'],'MVISC')
    >> tvisc = vint(mvisc)
    >> tvisc['DATA'][0,-1] # total viscous torque at first time step in Nm
    2.0986965
    >> mvisc2 = vder(tvisc)
    >> np.all(np.isclose(mvisc2['DATA'][0,:],mvisc['DATA'][0,:]))
    True

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = vint(v, dvol)
        return dnew

    if dvol == None:
        if 'DVOL' in d:
            dvol = d['DVOL']
        elif 'MDS' in d:
            dvol = d['MDS']['OUTPUTS']['TWO_D']['DVOL']
        else:
            dvol = OMFITmdsValue(d.server, d.treename, d.shot, TDI='DVOL')
    elif isinstance(dvol, OMFITmdsValue):
        dvol = OMFITtranspData(dvol)
    # if type(d)==OMFITmdsValue:
    #    d = OMFITtranspData(d)
    dvol = set_grid(dvol, 'X')  # just in case it was changed and given
    dint = set_grid(d, 'XB')
    ddv = set_grid(d, 'X')['DATA'] * dvol['DATA']
    dint['DATA'] = np.cumsum(ddv, axis=1)

    # clean up metadata
    dint['LABEL'] = 'int({:},dV)'.format(dint['LABEL'])
    dint['RPLABEL'] = dint['RPLABEL'].replace('DENSITY', '')
    dint['UNITS'] = dint['UNITS'] + '*CM**3'
    dint['UNITS'] = dint['UNITS'].replace('/CM3*CM**3', '')  # TRANSP conventions
    dint['UNITS'] = dint['UNITS'].replace('/Cm3*CM**3', '')  # TRANSP conventions
    dint['UNITS'] = dint['UNITS'].replace('/CM3/SEC*CM**3', '/SEC')  # TRANSP conventions

    return dint


OMFITtranspData.vint = vint
OMFITtranspMultigraph.vint = vint


def vder(d, dvol=None):
    """
    Derivative with respect to volume for TRANSP variables consistent with
    TRANSP finite differencing methods.

    See Solomon's unvolint.pro

    :param d: OMFITtranspData object from the mds+ TRANSP tree.
    :type d: OMFITtranspData
    :param dvol: OMFITtranspData object 'dvol' from the mds TRANSP tree.
        If None, will be taken from the Data's MDStree.
    :type dvol: OMFITtranspData or None

    :return: dy/dV OMFITtransData object on zone-centered grid.

    **Example:**
    Assuming the root is an OMFIT TRANSP module with a loaded run.
    >> mvisc = OMFITtranspData(root['OUTPUTS']['TRANSP_OUTPUT'],'MVISC')
    >> tvisc = vint(mvisc)
    >> tvisc['DATA'][0,-1] # total viscous torque at first time step in Nm
    2.0986965
    >> mvisc2 = vder(tvisc)
    >> np.all(np.isclose(mvisc2['DATA'][0,:],mvisc['DATA'][0,:]))
    True

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = vder(v, dvol)
        return dnew

    if dvol == None:
        dvol = d['DVOL']
    elif isinstance(dvol, OMFITmdsValue):
        dvol = OMFITtranspData(dvol)
    if isinstance(d, OMFITmdsValue):
        d = OMFITtranspData(d)
    dvol = set_grid(dvol, 'X')  # just in case it was changed and given
    dder = set_grid(d, 'X')
    y = set_grid(d, 'XB')['DATA']
    y = np.hstack([np.zeros([len(dder['DIM_OF'][1]), 1]), y])
    dder['DATA'] = np.diff(y, axis=1) / dvol['DATA']

    # clean up metadata
    dder['LABEL'] = 'd{:}/dV)'.format(dder['LABEL'])
    dder['UNITS'] = dder['UNITS'] + '/CM**3'
    dder['UNITS'] = dder['UNITS'].replace('CM3/CM**3', '')  # TRANSP conventions
    dder['UNITS'] = dder['UNITS'].replace('Cm3/CM**3', '')  # TRANSP conventions

    return dder


OMFITtranspData.vder = vder
OMFITtranspMultigraph.vder = vder


def sint(d, darea=None):
    """
    Surface integrate a TRANSP OMFITmdsValue object.
    Currently only available for objects from the TRANSP tree.

    :param d: OMFITtranspData object from the mds+ TRANSP OUTPUTS.TWO_D tree.
    :type d: OMFITtranspData

    :param darea: OMFITtranspData object 'darea' from the mds TRANSP tree.
        If None, will be taken from Data's MDStree.
    :type darea: OMFITtranspData or None

    **Example:**

    mds = OMFITmds('DIII-D','transp',1633030101)
    cur = OMFITtranspData(mds,'CUR')
    da = OMFITtranspData(mds,'DAREA')
    curi = cur.sint(darea=da)
    print(curi['DATA'][0,-1])
    pcur = OMFITtranspData(mds,'PCUR')
    print(pcur['DATA'][0])
    -> 1.16626e+06
    -> 1.16626e+06

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = sint(v, darea)
        return dnew

    if darea == None:
        if 'DAREA' in d:
            darea = d['DAREA']
        elif 'MDS' in d:
            darea = d['MDS']['OUTPUTS']['TWO_D']['DAREA']
        else:
            darea = OMFITmdsValue(d.server, d.treename, d.shot, TDI='DAREA')
    elif isinstance(darea, OMFITmdsValue):
        darea = OMFITtranspData(darea)
    # if type(d)==OMFITmdsValue:
    #    d = OMFITtranspData(d)
    darea = set_grid(darea, 'X')  # just in case it was changed and given
    dint = set_grid(d, 'XB')
    dda = set_grid(d, 'X')['DATA'] * darea['DATA']
    dint['DATA'] = np.cumsum(dda, axis=1)

    # clean up metadata
    dint['LABEL'] = 'int({:},dA)'.format(dint['LABEL'])
    dint['RPLABEL'] = dint['RPLABEL'].replace('DENSITY', '')
    dint['UNITS'] = dint['UNITS'] + '*CM**2'
    dint['UNITS'] = dint['UNITS'].replace('/CM2*CM**2', '')  # TRANSP conventions
    dint['UNITS'] = dint['UNITS'].replace('/Cm2*CM**2', '')  # TRANSP conventions

    return dint


OMFITtranspData.sint = sint
OMFITtranspMultigraph.sint = sint


def set_grid(d, zone='X', dvol=None):
    """
    Interpolate 2D TRANSP data to rho grid zone-boundary or zone-centered values.

    :param d: OMFITtranspData object from the mds+ TRANSP tree.

    :param zone: ``'XB'`` for zone-boundary rho, ``'X'`` for zone-centered. ``'V'`` or ``'VB'`` for volume.
    :type zone: str
    :param dvol: OMFITtranspData object 'dvol' from the mds+ TRANSP tree.
            If None, will be taken from Data's MDStree.

    return: OMFITtranspData object on the specified rho grid

    **Example:**
    Assuming the root is an OMFIT TRANSP module with a loaded run.
    >> mvisc = OMFITtranspData(root['OUTPUTS']['TRANSP_OUTPUT'],'MVISC')
    >> mvisc['XAXIS']
    'X'
    >> print(mvisc['DIM_OF'][0][:3])
    [ 0.01  0.03  0.05]
    >> mviscb = set_grid(mvisc,'XB')
    >> mviscb['XAXIS']
    'XB'
    >> print(mviscb['DIM_OF'][0][:3])
    [ 0.02  0.04  0.06]

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = set_grid(v, zone, dvol)
        return dnew

    if dvol == None and 'V' in zone:
        dvol = d['DVOL']
    elif isinstance(dvol, OMFITmdsValue):
        dvol = OMFITtranspData(dvol)
    if isinstance(d, OMFITmdsValue):
        d = OMFITtranspData(d)
    zone = zone.upper()
    db = copy.copy(d)

    # return a copy if no change
    if d['XAXIS'] == zone:
        return db

    # native transp rho grid center or boundary
    if zone in ['X', 'XB']:
        if 'MDS' in d:
            xnew = d['MDS']['TRANSP_OUT'][zone].data()
        else:
            xnew = d['CDF'][zone]['data']
        y = d['DATA']  # .ravel()
        x = d['DIM_OF'][0][0, :]  # .ravel()
        t = d['DIM_OF'][1][:, 0]  # .ravel()
        # f = LinearNDInterpolator(zip(x,t),y)
        # db['DATA'] = f(zip(xnew.ravel(),t.ravel())).reshape(t.shape)
        f = RectBivariateSpline(x, t, y.T)  # assume regular grid
        db['DATA'] = f(xnew[0, :], t).T
        db['DIM_OF'][0] = xnew
    # volume grid center or boundary for easy integration of densities
    elif zone in ['V', 'VB']:
        db = set_grid(d, zone.replace('V', 'X'))
        v = set_grid(xint(dvol), zone.replace('V', 'X'))
        db['DIM_OF'][0] = v['DATA']
    else:
        raise ValueError('Valid zones are "X", "XB", "V", or "VB".')
    db['XAXIS'] = zone
    return db


OMFITtranspData.set_grid = set_grid
OMFITtranspMultigraph.set_grid = set_grid


def xdiff(d):
    """
    TRANSP convention of simple finite difference along rho (axis=0), including
    switching of centered/boundary grids.

    :param d: OMFITtranspData object from the mds+ TRANSP tree.

    :return: OMFITtranspData differenced on the other rho grid (``'X'`` vs ``'XB'``)

    **Example:**

    >> x = Data(['x','TRANSP'],1470670204)
    >> dx = xdiff(x)
    >> print(dx['XAXIS'])
    XB
    >> print(dx.y[0,:3])
    [ 0.02  0.02  0.02]

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = xdiff(v)
        return dnew

    ddiff = copy.copy(d)
    # extend axis by 1 (back of forward)
    tn = list(d['DIM_OF'][1].T)
    tn.insert(0, d['DIM_OF'][1][:, 0])
    xn = list(d['DIM_OF'][0].T)
    if d['XAXIS'] == 'X':
        xn.insert(len(xn), xn[-1] + (xn[-1] - xn[-2]))
        ddiff = set_grid(ddiff, 'XB')
    elif d['XAXIS'] == 'XB':
        xn.insert(0, xn[0] - (xn[1] - xn[0]))
        ddiff = set_grid(ddiff, 'X')
    else:
        raise ValueError('Data object must have xaxis attribute "X" or "XB"')
    # data on extended axis
    y = d['DATA'].ravel()
    x = d['DIM_OF'][0].ravel()
    t = d['DIM_OF'][1].ravel()
    fl = LinearNDInterpolator(list(zip(x, t)), y)
    fn = NearestNDInterpolator(list(zip(x, t)), y)
    dn = fl(list(zip(np.array(xn).ravel(), np.array(tn).ravel()))).reshape(np.shape(xn))
    dn[np.isnan(dn)] = fn(list(zip(np.array(xn)[np.isnan(dn)].ravel(), np.array(tn)[np.isnan(dn)].ravel())))
    # simple finite differences
    dy = np.diff(dn, axis=0).T
    # new data
    ddiff['DATA'] = dy
    # clean up metadata
    ddiff['LABEL'] = 'd({:})'.format(d['LABEL'])
    ddiff['RPLABEL'] = 'Delta {:})'.format(d['RPLABEL'])

    return ddiff


OMFITtranspData.xdiff = xdiff
OMFITtranspMultigraph.xdiff = xdiff


def xder(d):
    """
    TRANSP style differentiation in rho.

    :param d: OMFITtranspData object from the TRANSP tree.

    :return: dy/drho OMFITtranspData object on the other rho grid.

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = xder(v)
        return dnew

    dx = xdiff(OMFITtranspData(d['MDS'], d['XAXIS']))
    dder = xdiff(d)
    dder['DATA'] /= dx['DATA']

    # clean up metadata
    dder['LABEL'] = '{:}/drho'.format(dder['LABEL'])

    return dder


OMFITtranspData.xder = xder
OMFITtranspMultigraph.xder = xder


def xint(d):
    """
    TRANSP style integration in rho. UNVALIDATED.

    :param d: OMFITtranspData object from the TRANSP tree.

    :return: dy/drho OMFITtranspData object on the other rho grid.

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = xint(v)
        return dnew

    dint = copy.copy(d)
    dint['DATA'] = integrate.cumtrapz(dint['DATA'], x=dint['DIM_OF'][0], axis=1, initial=0)

    # clean up metadata
    dint['LABEL'] = 'int({:},drho)'.format(dint['LABEL'])

    return dint


OMFITtranspData.xint = xint
OMFITtranspMultigraph.xint = xint


def aol(d, rmnmp=None):
    """
    Normalized inverse scale length a/Lx with derivative with respect to
    midplane minor radius "r".  The equation is aol = -(a/X)dX/dr

    :param d: OMFITtranspData object from the TRANSP tree.
    :param rmnmp: OMFITtranspData "RMNMP"

    :return:

    """
    if isinstance(d, OMFITtranspMultigraph):
        dnew = copy.copy(d)
        for k, v in list(dnew['CONTENT'].items()):
            dnew['CONTENT'][k] = aol(v, rmnmp)
        return dnew

    if rmnmp == None:
        if 'RMNMP' in d:
            rmnmp = d['RMNMP']
        elif 'MDS' in d:
            rmnmp = d['MDS']['OUTPUTS']['TWO_D']['RMNMP']
        else:
            rmnmp = OMFITmdsValue(d.server, d.treename, d.shot, TDI='RMNMP')
    elif isinstance(rmnmp, OMFITmdsValue):
        rmnmp = OMFITtranspData(rmnmp)

    a = rmnmp['DATA'][:, -1]
    dr = np.zeros(d['DATA'].shape)
    for i in range(d['DATA'].shape[0]):
        dr[i, :] = deriv(rmnmp['DATA'][i, :], d['DATA'][i, :])
    dder = copy.copy(d)
    dder['DATA'] = -1.0 * (a[:, np.newaxis] / d['DATA']) * dr

    # clean up metadata
    dder['LABEL'] = 'a/L_{:})'.format(dder['LABEL'])
    dder['UNITS'] = '-'
    return dder


OMFITtranspData.aol = aol
OMFITtranspMultigraph.aol = aol


def tavg(d, time=None, avgtime=None):
    """
    Time average data

    :param d: OMFITtranspData object from the TRANSP tree.
    :param time: Center time in seconds
    :param avgtime: Averaging window (+/-) in seconds

    :return: For 1D input uncertainty uarray of the data time avarge and standard deviation.
             For 2D input uncertainty uarray of the profile with time avarge and standard deviation.

    **Example:**
    Assuming data in root['OUTPUTS']['TRANSP_OUTPUT']

    time = 2.0
    avgtime = 0.1
    # 1D
    tr_neutt = OMFITtranspData(root['OUTPUTS']['TRANSP_OUTPUT'], 'NEUTT')
    tr_neutt_tavg = tr_neutt.tavg(time=time, avgtime=avgtime)
    tr_neutt.plot()
    ax = gca()
    uband([time-avgtime, time+avgtime],np.repeat(tr_neutt_tavg,2),ax=ax)
    ax.text(time,nominal_values(tr_neutt_tavg)+std_devs(tr_neutt_tavg),
            '{0:0.3g}+/-{1:0.3g}'.format(nominal_values(tr_neutt_tavg),std_devs(tr_neutt_tavg)))

    # 2D
    tr_ne = OMFITtranspData(root['OUTPUTS']['TRANSP_OUTPUT'],'NE')
    tr_ne_tavg = tr_ne.tavg(time=time, avgtime=avgtime)
    figure()
    tr_ne.plot(slice_axis=1,slice_at=time)
    ax = gca()
    uband(tr_ne['DIM_OF'][0][0,:],tr_ne_tavg,ax=ax)

    """

    if time is None or avgtime is None:
        raise (OMFITexception("Must set time and average time"))

    # Determine if 1D or 2D data
    if len(d['DIM_OF']) == 1 and d['UNITS_DIM_OF'][0] == 's':
        # 1D
        idx = np.where((d['DIM_OF'][0] >= time - avgtime) * (d['DIM_OF'][0] <= time + avgtime))[0]
        if len(idx):
            ud = uarray(np.mean(d['DATA'][idx]), np.std(d['DATA'][idx]))
        else:
            raise (OMFITexception("Averaging window empty."))
    elif len(d['DIM_OF']) == 2 and d['UNITS_DIM_OF'][1] == 's':
        # 2D
        idx = np.where((d['DIM_OF'][1][:, 0] >= time - avgtime) * (d['DIM_OF'][1][:, 0] <= time + avgtime))[0]
        if len(idx):
            ud = uarray(np.mean(d['DATA'][idx, :], axis=0), np.std(d['DATA'][idx, :], axis=0))
        else:
            raise (OMFITexception("Averaging window empty."))
    else:
        raise (OMFITexception("Unrecognized dimensions of quantity to average."))

    return ud


OMFITtranspData.tavg = tavg


############################################
if '__main__' == __name__:
    test_classes_main_header()

    # This is not a real TRANSP output file, but the class should be able to finish __init__() with any .nc sample.
    # This is just a limited initialization test, not a detailed test of all class features.
    test_nc_file = os.sep.join([OMFITsrc, '..', 'samples', 'bmn.nc'])
    class_instance = OMFITtranspOutput(test_nc_file)
    class_instance.load()



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_matrix import OMFITmatrix
from omfit_classes.omfit_eqdsk import OMFITgeqdsk

from traceback import extract_stack
from warnings import filterwarnings
import numpy as np
import xarray as xr
import pandas as pd

__all__ = ['OMFITtrip3Dcoil', 'OMFITtrip3Dlines', 'OMFITtrip3Dhits', 'OMFITtrip3Dstart', 'OMFITtrip3Dprobeg']


def foot_on_line(xp, yp, x1, x2, y1, y2):
    r"""
    Given a point (P) and a segment (1), (2), compute the foot (F) on the line and the point-to-line distance d.

             (P) \                  (P) \
              :  |                   :  |
              :  | d                 :  | d
              :  |                   :  |
              :  /                   :  /
        (1)==(F)===========(2)      (F)------(1)=======(2)

    An array of points P is allowed, but there can only be one line segment.

    :param xp: numeric
        X-coordinate of point P or array of X-coordinates of several points

    :param yp: numeric
        Y-coordinate of point P or array of Y-coordinates of several points (must be scalar or match the shape of xp)

    :param x1: float
        X-coordinate of one end of the line segment (point 1)

    :param x2: float
        X-coordinate of the other end of the line segment (point 2)

    :param y1: float
        Y-coordinate of one end of the line segment (point 1)

    :param y2: float
        Y-coordinate of the other end of the line segment (point 2)

    :return: tuple
        X-coordinate(s) of the foot point as a float or array (matching the shape of xp or yp)
        Y-coordinate(s) of the foot point as a float or array (matching the shape of xp or yp)
        Distance between the point(s) P and the line as a float or array (matching the shape of xp or yp)
    """

    if x1 == x2 and y1 == y2:
        raise ValueError('(x1,y1) == (x2,y2)')

    elif x1 == x2:  # vertical

        xf = x1 * xp**0
        yf = yp * xp**0

    elif y1 == y2:  # horizontal

        xf = xp * xp**0
        yf = y1 * xp**0

    else:  # diagonal

        # (1) and (2)
        mm = (y2 - y1) / (x2 - x1)
        qq = y1 - mm * x1

        # (p) and (f)
        nn = -1.0 / mm
        pp = yp - nn * xp

        # foot
        xf = (pp - qq) / (mm - nn)
        yf = nn * (xf - xp) + yp

    # distance
    dd = np.hypot(xf - xp, yf - yp)

    return xf, yf, dd


def sshape(d):
    return 'x'.join([str(s) for s in np.shape(d)])


# Created by wuwen at 2015/01/22 15:50
class OMFITtrip3Dcoil(SortedDict, OMFITascii):
    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    def __tree_repr__(self):

        myrepr = 'FILE: %s   (%s' % (os.path.split(self.filename)[1], sizeof_fmt(self.filename, ' '))
        if not self.dynaLoad:
            myrepr += ', {} loops'.format(len(self) - 2)
        myrepr += ')'
        return myrepr, []

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.readlines()

        self['device'] = lines[0].strip()
        self['coil'] = lines[1].strip()

        nloops = int(lines[2].split()[0])

        k = 3
        for l in range(nloops):
            nitems = int(lines[k].split()[0])
            titem = lines[k].split()[1]
            k += 1
            loop = SortedDict()
            for s in range(nitems):
                loop[s] = [float(i) for i in lines[k].split()]
                k += 1
            self[l, titem] = loop

    @dynaSave
    def save(self):
        with open(self.filename, 'w') as f:
            f.write(self['device'].strip() + '\n')
            f.write(self['coil'].strip() + '\n')
            f.write('%3d   loops\n' % (len(self) - 2))
            for k in list(self.keys())[2:]:
                f.write('%3d   %s\n' % (len(self[k]), k[1]))
                for c in list(self[k].keys()):
                    f.write(('  % f' * len(self[k][c]) + '\n') % tuple(self[k][c]))


# Created by Trevisan at 05 Dec 2016  15:49
class OMFITtrip3Dlines(OMFITmatrix):

    """
    The OMFITtrip3Dlines class parses and handles the TRIP3D 'lines.out' output file.
    A self-described xarray object is stored under self['data'].
    """

    # file descriptors
    columns = {
        2: ['psimin', 'psimax'],
        # 8: ['phi','rho','ith','the','bmag','brot','psin','mlen'],
        9: ['phi', 'rho', 'ith', 'the', 'brho', 'bthe', 'bphi', 'mlen', 'psin'],
    }
    for c in [9]:
        columns[c + 2] = list(columns[c])
        columns[c + 2].extend(['rr', 'zz'])
        columns[c + 4] = list(columns[c + 2])
        columns[c + 4].extend(['psimin', 'psimax'])

    def __getitem__(self, key):

        try:
            return SortedDict.__getitem__(self, key)
        except Exception:
            if key == 'data':
                raise
        try:
            return self['data'].loc[:, :, key].values
        except Exception:
            return self['data'].attrs[key]

    def __mul__(self, other, quiet=False):

        # sanity checks
        if not isinstance(other, self.__class__):
            raise TypeError('class mismatch: {} vs {}.'.format(other.__class__.__name__, self.__class__.__name__))
        if self['data'].shape[0] != other['data'].shape[0]:
            raise ValueError(
                'shape mismatch: {} vs {} (full dims: {}, {}).'.format(
                    self['data'].shape[0], other['data'].shape[0], self['data'].shape, other['data'].shape
                )
            )

        # psiN arrays
        self['data'].loc[:, :, 'psimin'] = (self['psin'].transpose() ** 0 * np.nanmin(other['data'][:, 0, :], axis=-1)).transpose()
        self['data'].loc[:, :, 'psimax'] = (self['psin'].transpose() ** 0 * np.nanmax(other['data'][:, 0, :], axis=-1)).transpose()

        if not quiet:
            _print(self, 'updated psimin and psimax.')

        return self

    @dynaLoad
    def load(self, bin=None, zip=None, geqdsk=None, verbose=False, quiet=False):

        p = False
        t = tstart = time.time()

        # read file
        bin, zip = OMFITmatrix.load(
            self, bin=bin, zip=zip, header=None, comment='%', delim_whitespace=True, index_col=False, na_filter=False, engine='c'
        )

        if not quiet and verbose:
            p = _print(self, '%.2f s, %s, read from file.' % (time.time() - tstart, sshape(self.get('data', None))), p=p)

        if not bin:

            # reshape to 3D
            tstart = time.time()
            self.to3d()
            if not quiet and verbose:
                p = _print(self, '%.2f s, %s, reshaped to 3D.' % (time.time() - tstart, sshape(self.get('data', None))), p=p)

            if np.shape(self['data'])[-1] in [8, 9]:

                # add attributes
                tstart = time.time()
                stat = self.addattrs(geqdsk=geqdsk)
                if not quiet and verbose:
                    if stat is None:
                        p = _print(self, '%.2f s, did not read gEQDSK data.' % (time.time() - tstart), tag='w', p=p)
                    else:
                        p = _print(self, '%.2f s, read gEQDSK data.' % (time.time() - tstart), p=p)

                # add columns
                tstart = time.time()
                self.addcols()
                if not quiet and verbose:
                    p = _print(self, '%.2f s, %s, added columns.' % (time.time() - tstart, sshape(self['data'])), p=p)

        # summary
        if not quiet:
            if zip:
                z = 'zlibbed' if bin else 'gzipped'
            else:
                z = 'plain'
            p = _print(
                self, '%.2f s, %s, loaded %s %s.' % (time.time() - t, sizeof_fmt(self.filename, ' '), z, 'NetCDF' if bin else 'ASCII'), p=p
            )

    def to3d(self):

        old = self['data'].values
        s = np.shape(old)
        if s[-1] == 2:
            lidx = np.array(list(range(s[0]))).astype(int)
            llen = lidx**0
        elif s[-1] in [8, 9]:
            lidx = np.where(old[:, -2] == 0)[0].astype(int)
            llen = np.diff(np.append(lidx, s[-2])).astype(int)
        else:
            raise ValueError('unexpected shape: ' + str(s))

        new = np.nan * np.zeros((np.size(lidx), np.max(np.append(llen, 1)), s[-1]))
        for idx, (at, len) in enumerate(zip(lidx, llen)):
            new[idx, :len, :] = old[at : at + len, :]

        self.toxr(new)

    def addattrs(self, geqdsk=None):

        if geqdsk is None:
            try:
                geqdsk = eval(treeLocation(self)[-3])['INPUTS']['gEQDSK']
            except Exception:
                pass
        try:
            rmax = geqdsk['RMAXIS']
            zmax = geqdsk['ZMAXIS']
            rlim = geqdsk['RLIM']
            zlim = geqdsk['ZLIM']
            stat = True
        except Exception:
            rmax = np.nan
            zmax = np.nan
            rlim = np.nan
            zlim = np.nan
            stat = None
        finally:
            self['data'].attrs['rmax'] = rmax
            self['data'].attrs['zmax'] = zmax
            self['data'].attrs['rlim'] = rlim
            self['data'].attrs['zlim'] = zlim
            return stat

    def addcols(self):

        filterwarnings('ignore', r'All-NaN (slice|axis) encountered')

        old = self['data'].values
        s = np.shape(old)
        try:
            rmax = self['rmax']
            zmax = self['zmax']
            if np.isnan(rmax) or np.isnan(zmax):
                raise ValueError
        except Exception:
            rmax = zmax = 0.0

        new = np.nan * np.zeros((s[0], s[1], s[2] + 4))

        new[:, :, :-4] = old[:, :, :]
        new[:, :, s[2] + 0] = rmax + self['rho'] * np.cos(np.deg2rad(self['the']))
        new[:, :, s[2] + 1] = zmax + self['rho'] * np.sin(np.deg2rad(self['the']))
        new[:, :, s[2] + 2] = (self['psin'].transpose() ** 0 * np.nanmin(self['psin'], axis=-1)).transpose()
        new[:, :, s[2] + 3] = (self['psin'].transpose() ** 0 * np.nanmax(self['psin'], axis=-1)).transpose()

        self.toxr(new)

    def toxr(self, data, attrs=None):

        s = np.shape(data)
        l = list(range(s[0]))
        p = list(range(s[1]))
        c = self.columns[s[2]]
        if attrs is None:
            attrs = self['data'].attrs

        self['data'] = xr.DataArray(data, name='lines', dims=['L', 'P', 'C'], coords=[l, p, c], attrs=attrs)

    @dynaSave
    def save(self, zip=True, quiet=False):

        t = time.time()

        bin, zip = OMFITmatrix.save(self, bin=True, zip=zip)

        if not quiet:
            if zip:
                z = 'zlibbed' if bin else 'gzipped'
            else:
                z = 'plain'
            _print(self, '%.2f s, %s, saved %s %s.' % (time.time() - t, sizeof_fmt(self.filename, ' '), z, 'NetCDF' if bin else 'ASCII'))

    def prep(self, type, prefix=True, quiet=False):

        # disable method for non-line instances
        if np.shape(self['data'])[-1] == 2:
            return

        # prep type
        if type not in ['poincare', 'contour', 'control']:
            raise TypeError('unexpected value for prep type')

        # load before starting the stopwatch
        if self.dynaLoad:
            self.load(prefix=prefix, quiet=quiet)

        # stopwatch and summary
        tstart = time.time()

        # check data
        if '__%s__' % type in self:
            return

        # prepare data
        data = self['data']
        if type == 'poincare':

            la = r"poloidal angle, $\vartheta$ [deg]"
            aa = self['the'].copy()
            numsi, numsj = np.where(np.isnan(aa) == False)
            high = np.where(aa[numsi, numsj] > 180)[0]
            aa[numsi[high], numsj[high]] -= 360
            lb = r"normalized poloidal flux, $\Psi_N$"
            bb = self['psin'].copy()

        elif type == 'contour':

            la = r"$R - R_{ax}$ [m]"
            aa = self['rr'].copy()
            lb = r"$Z - Z_{ax}$ [m]"
            bb = self['zz'].copy()

        elif type == 'control':

            la = r"normalized toroidal angle, $\phi \% 360$ [deg]"
            aa = self['phi'] % 360
            lb = r"toroidal angle, $\phi$ [deg]"
            bb = self['phi'].copy()

        # initial psin
        lc = r"initial $\Psi_N$"
        cc = np.transpose(np.transpose(self['psin']) ** 0 * self['psin'][:, 0])

        # minimum psin
        ld = r"min $\Psi_N$"
        dd = self['psimin'].copy()

        # store data
        self['__%s__' % type] = np.transpose([np.transpose(aa), np.transpose(bb)])
        self['__colors__'] = np.transpose([np.transpose(cc), np.transpose(dd)])
        labels = self.setdefault('__labels__', {})
        labels[type] = [la, lb]
        labels['colors'] = [lc, ld]

        # stopwatch and summary
        if not quiet:
            _print(self, '%s: prepared data in %.3f s.' % (type.ljust(8), time.time() - tstart), prefix=prefix)

    def doplot(
        self,
        type,
        geqdsk=None,
        lines=(None, None, None),
        points=(None, None, None),
        limx=(None, None),
        limy=(None, None),
        col='min',
        cbar=True,
        prefix=True,
        quiet=False,
        **kw,
    ):

        # disable method for non-line instances
        if np.shape(self['data'])[-1] == 2:
            return

        # printed
        p = False

        # plot type
        if type not in ['poincare', 'contour', 'control']:
            raise TypeError('unexpected value for plot type: %s' % type)

        # color type
        if col not in ['min', 'ini']:
            raise TypeError('unexpected value for color: %s' % col)

        # prepare data
        self.prep(type=type, prefix=prefix, quiet=quiet)

        # retrieve data
        aa = self['__%s__' % type][:, :, 0]
        bb = self['__%s__' % type][:, :, 1]
        [la, lb] = self['__labels__'][type][:2]
        if col == 'ini':
            c = 0
        elif col == 'min':
            c = 1
        cc = self['__colors__'][:, :, c]
        lc = self['__labels__']['colors'][c]

        # stopwatch and summary
        tstart = time.time()

        # line slicing
        if isinstance(lines, tuple):
            lsli = slice(*lines)
            if not quiet and lines != (None, None, None):
                p = _print(self, '%s: slicing lines  as %s.' % (type.ljust(8), repr(lines)), prefix=prefix, p=p)
        elif isinstance(lines, list) and np.size(lines):
            lsli = lines
            if not quiet:
                p = _print(self, '%s: selecting %d lines.' % (type.ljust(8), np.size(lines)), prefix=prefix, p=p)
        else:
            raise TypeError('unexpected value for lines')

        # point slicing
        if isinstance(points, tuple):
            psli = slice(*points)
            if not quiet and points != (None, None, None):
                p = _print(self, '%s: slicing points as %s.' % (type.ljust(8), repr(points)), prefix=prefix, p=p)
        elif isinstance(points, list) and np.size(points):
            psli = points
            if not quiet:
                p = _print(self, '%s: selecting %d points.' % (type.ljust(8), np.size(points)), prefix=prefix, p=p)
        else:
            raise TypeError('unexpected value for points')

        # default boundaries
        kw.setdefault('vmin', np.nanmin(cc[lsli, psli].reshape(-1)))
        kw.setdefault('vmax', np.nanmax(cc[lsli, psli].reshape(-1)))

        # default colormap
        if 'cmap' not in kw:

            intmid = int(1e4)
            inthig = int(intmid * kw['vmin'])
            intlow = int(intmid * kw['vmax'])

            higbar = intmid > inthig
            lowbar = intlow > intmid
            if higbar:
                hi = pyplot.cm.autumn(np.linspace(0, 1, intmid - inthig))
            if lowbar:
                lo = pyplot.cm.winter_r(np.linspace(0, 1, intlow - intmid))
            if higbar and lowbar:
                mi = int((intlow - inthig) // 100) * [[0, 0, 0, 1]]
                bar = np.vstack((hi, mi, lo))
            else:
                bar = hi if higbar else lo

            kw['cmap'] = matplotlib.colors.LinearSegmentedColormap.from_list('cmapall', bar)
            kw['vmin'] = 1.0 * inthig / intmid
            kw['vmax'] = 1.0 * intlow / intmid

        # contour plot
        refa = 0
        refb = 0
        if type == 'contour':

            pyplot.gca().set_aspect('equal')

            # try to locate the gEQDSK in INPUTS
            if geqdsk is None:
                try:
                    geqdsk = eval(treeLocation(self)[-3])['INPUTS']['gEQDSK']
                    if not isinstance(geqdsk, OMFITgeqdsk):
                        raise TypeError
                except Exception:
                    geqdsk = None
            try:
                refa = self['rmax']
                refb = self['zmax']
                la = r'$R$ [m]'
                lb = r'$Z$ [m]'
                pyplot.plot(self['rlim'], self['zlim'], 'k-')
            except Exception:
                if not quiet:
                    p = _print(self, '%s: R/ZLIM will not be plotted.' % type.ljust(8), tag='w', prefix=prefix, p=p)

            pyplot.plot(refa, refb, 'k+')

        # plot
        kw.setdefault('s', 1)
        kw.setdefault('edgecolor', '')
        kw.setdefault('rasterized', True)
        scatter(aa[lsli, psli], bb[lsli, psli], c=cc[lsli, psli], **kw)

        # set limits
        if limx != (None, None):
            pyplot.xlim(limx)
        elif type == 'poincare':
            pyplot.xlim(-180, 180)
        if limy != (None, None):
            pyplot.ylim(limy)

        # set labels
        pyplot.xlabel(la)
        pyplot.ylabel(lb)

        if cbar:
            cb = colorbar()
            cb.ax.set_title(lc)
            if kw['vmin'] < 1.0 and 1.0 < kw['vmax']:
                try:
                    ticks = np.around(np.linspace(kw['vmin'], kw['vmax'], 7), decimals=3)
                    (tisep,) = np.where(abs(ticks - 1) == min(abs(ticks - 1)))
                    ticks += 1 - ticks[tisep]
                    cb.set_ticks(ticks)
                except Exception:
                    if not quiet:
                        p = _print(self, '%s: could not set colorbar ticks.' % type.ljust(8), tag='w', prefix=prefix, p=p)

        # stopwatch and summary
        tstop = time.time()
        num = np.size(aa[lsli, psli]) - np.sum(np.isnan(aa[lsli, psli]))
        if not quiet:
            p = _print(self, '{}: plotted {:,d} points in {:.3f} s.'.format(type.ljust(8), num, tstop - tstart), prefix=prefix, p=p)

    def plot(self, type='summary', **kw):

        # disable method for non-line instances
        if np.shape(self['data'])[-1] == 2:
            return

        # prefix
        me = self.__class__.__name__ + '.plot'
        blankme = ' ' * len(me)

        # quiet keyword
        quiet = kw.setdefault('quiet', False)

        # plot type
        if type in ['poincare', 'contour', 'control']:

            self.doplot(type=type, **kw)
            return

        elif type != 'summary':

            raise TypeError('unexpected value for plot type: %s' % type)

        # prepare data
        p = False
        for t in ['poincare', 'contour', 'control']:
            if '__%s__' % t in self:
                continue
            self.prep(type=t, prefix=blankme if p else True, quiet=quiet)
            p = True

        # default slicing
        p = False
        if 'lines' not in kw and 'points' not in kw:

            # line slicing: toroidal planes
            kw['lines'] = (None, None, None)
            (planes,) = np.where(self['data'][:, 0, 0] > self['data'][0, 0, 0])
            if len(planes):
                stop = planes[0]
                kw['lines'] = (None, stop, None)
                if not quiet:
                    p = _print(
                        self,
                        '%s: detected %d starting planes with step of %.1f deg.'
                        % (type.ljust(8), np.shape(self['data'])[0] / stop, self['data'][stop, 0, 0]),
                        p=p,
                    )
            else:
                if not quiet:
                    p = _print(self, '%s: detected 1 starting plane at %.1f deg.' % (type.ljust(8), self['data'][0, 0, 0]), p=p)

            # point slicing: poincare planes
            kw['points'] = (None, None, None)
            step = 0
            for idx in [0, int(np.shape(self['data'])[1] // 2)]:
                try:
                    diff = abs(self['data'][0, idx + 1, 0] - self['data'][0, idx, 0])
                    step = int(round(360.0 / diff))
                    break
                except Exception:
                    pass
            if step == 0:
                if not quiet:
                    p = _print(self, '%s: could not detect toroidal step.' % type.ljust(8), tag='w', p=p)
            elif step == 1:
                if not quiet:
                    p = _print(self, '%s: detected 1 poincare plane at %.1f deg.' % (type.ljust(8), self['data'][0, idx, 0]), p=p)
            else:
                kw['points'] = (None, None, step)
                if not quiet:
                    p = _print(self, '%s: detected %d poincare planes with step of %.1f deg.' % (type.ljust(8), step, round(diff)), p=p)

        else:

            kw.setdefault('points', (None, None, None))
            kw.setdefault('lines', (None, None, None))
            if not quiet:
                p = _print(self, '%s: overriding default slicing.' % type.ljust(8), p=p)

        # colormaps
        gg = matplotlib.colors.LinearSegmentedColormap.from_list('g', 2 * [[0.7, 0.7, 0.7, 1]])
        kk = matplotlib.colors.LinearSegmentedColormap.from_list('k', 2 * [[0, 0, 0, 1]])

        # composite plot
        ioff()
        suptitle(treeLocation(self)[-1])

        # control plots
        pyplot.subplot(2, 2, 1)
        if 'points' in kw and kw['points'] != (None, None, None):

            kwa = kw.copy()
            kwa['points'] = (None, None, None)
            kwa['cmap'] = gg
            kwa.pop('quiet')
            kwa['cbar'] = False

            # downsample lines for performance
            if isinstance(kw['lines'], tuple):
                lsli = slice(*kw['lines'])
            elif isinstance(kw['lines'], list) and np.size(kw['lines']):
                lsli = kw['lines']
            lidx = list(range(np.shape(self['data'])[0]))[lsli]
            got = np.count_nonzero(~np.isnan(self['data'][lidx, :, 0]))
            want = 123456
            if got > want:
                step = got / want + int(got % want > 0)
                kwa['lines'] = lidx[::step]
                if not quiet:
                    p = _print(self, '{:s}: downsampled control for performance.'.format(type.ljust(8)), p=p)

            self.doplot(type='control', quiet=True, **kwa)

        kwb = kw.copy()
        kwb['s'] = 6
        kwb['cmap'] = kk
        kwb['cbar'] = False
        self.doplot(type='control', prefix=blankme, **kwb)

        # poincare plot
        pyplot.subplot(2, 2, 3)
        kwc = kw.copy()
        kwc['cbar'] = False
        self.doplot(type='poincare', prefix=blankme, **kwc)

        # contour plot
        pyplot.subplot(1, 2, 2)
        self.doplot(type='contour', prefix=blankme, **kw)

        # interactive plotting
        ion()
        draw()
        show()


# Created by Trevisan at 11 Aug 2017  16:03
class OMFITtrip3Dhits(OMFITmatrix):

    """
    The OMFITtrip3Dhits class parses and handles the TRIP3D 'hit.out' output file.
    A self-described xarray object is stored under self['data'].
    """

    def __getitem__(self, key):

        try:
            return SortedDict.__getitem__(self, key)
        except Exception:
            if key == 'data':
                raise
        try:
            return self['data'].loc[:, :, key].values
        except Exception:
            return self['data'].attrs[key]

    @dynaLoad
    def __invert__(self):

        s = copy.deepcopy(self)
        s['data'].attrs = self['data'].attrs.copy()
        for att in 'abc':
            s['data'].attrs.pop(att, None)
        s['data'].loc[:, 0, :] = self['data'].loc[:, 2, :].copy()
        s['data'].loc[:, 2, :] = self['data'].loc[:, 0, :].copy()
        s['data'].loc[:, :, 'mlen'] = self['data'].loc[:, :, 'mlen'].copy()
        s['data'][:, 1, :].data *= np.nan
        s.filename = s.link = OMFITascii('hit.inv').filename
        if self.readOnly:
            s.readOnly = False
            s.save()
            s.readOnly = True
        return s

    @dynaLoad
    def load(self, bin=None, zip=None, inverse=True, geqdsk=None, verbose=False, quiet=False):

        p = False
        t = tstart = time.time()

        # read file
        bin, zip = OMFITmatrix.load(self, bin=bin, zip=zip)

        if not quiet and verbose:
            p = _print(self, '%.2f s, %s, read from file.' % (time.time() - tstart, sshape(self['data'])), p=p)

        if not bin:

            # reshape to 3D
            tstart = time.time()
            self.to3d()
            if not quiet and verbose:
                p = _print(self, '%.2f s, %s, reshaped to 3D.' % (time.time() - tstart, sshape(self['data'])), p=p)

            # add attributes
            tstart = time.time()
            stat = self.addattrs(inverse=inverse, geqdsk=geqdsk)
            if not quiet and verbose:
                if stat is None:
                    p = _print(self, '%.2f s, did not read gEQDSK data.' % (time.time() - tstart), tag='w', p=p)
                else:
                    p = _print(self, '%.2f s, read %s gEQDSK data.' % (time.time() - tstart, 'inverted' if stat else 'normal'), p=p)

            # add columns
            tstart = time.time()
            self.addcols()
            if not quiet and verbose:
                p = _print(self, '%.2f s, %s, added columns.' % (time.time() - tstart, sshape(self['data'])), p=p)

        # summary
        if not quiet:
            if zip:
                z = 'zlibbed' if bin else 'gzipped'
            else:
                z = 'plain'
            p = _print(
                self, '%.2f s, %s, loaded %s %s.' % (time.time() - t, sizeof_fmt(self.filename, ' '), z, 'NetCDF' if bin else 'ASCII'), p=p
            )

    def to3d(self):

        old = self['data'].values
        s = np.shape(old)
        idx = old[:, 0].astype(int)
        cols = int((s[-1] - 1) // 3)

        new = np.nan * np.zeros((s[0], 3, cols))
        for p in range(3):
            new[:, p, :cols] = old[:, 1 + p * cols : 1 + (p + 1) * cols]

        self.toxr(new, attrs={'index': idx})

    def addattrs(self, inverse=True, geqdsk=None):

        if geqdsk is None:
            try:
                geqdsk = eval(treeLocation(self)[-3])['INPUTS']['gEQDSK']
            except Exception:
                pass
        try:
            rmax = geqdsk['RMAXIS']
            zmax = geqdsk['ZMAXIS']
            rlim = geqdsk['RLIM'][-1 if inverse else None :: -1 if inverse else None]
            zlim = geqdsk['ZLIM'][-1 if inverse else None :: -1 if inverse else None]
            slim = np.insert(np.cumsum(np.hypot(np.diff(rlim[:]), np.diff(zlim[:]))), 0, 0)
            stat = inverse
        except Exception:
            rmax = np.nan
            zmax = np.nan
            rlim = np.nan
            zlim = np.nan
            slim = np.nan
            stat = None
        finally:
            self['data'].attrs['rmax'] = rmax
            self['data'].attrs['zmax'] = zmax
            self['data'].attrs['rlim'] = rlim
            self['data'].attrs['zlim'] = zlim
            self['data'].attrs['slim'] = slim
            return stat

    def addcols(self):

        filterwarnings('ignore', r'All-NaN (slice|axis) encountered')

        old = self['data'].values
        s = np.shape(old)
        try:
            rmax = self['rmax']
            zmax = self['zmax']
            if np.isnan(rmax) or np.isnan(zmax):
                raise ValueError
        except Exception:
            rmax = zmax = 0.0

        new = np.nan * np.zeros((s[0], s[1], s[2] + 2))

        new[:, :, :-2] = old[:, :, :]
        new[:, :, s[2] + 0] = rmax + self['rho'] * np.cos(np.deg2rad(self['the']))
        new[:, :, s[2] + 1] = zmax + self['rho'] * np.sin(np.deg2rad(self['the']))

        self.toxr(new)

    def toxr(self, data, attrs=None):

        s = np.shape(data)
        l = list(range(s[0]))
        p = list(range(s[1]))
        c = OMFITtrip3Dlines.columns[s[2]]
        if attrs is None:
            attrs = self['data'].attrs

        self['data'] = xr.DataArray(data, name='hits', dims=['L', 'P', 'C'], coords=[l, p, c], attrs=attrs)

    @dynaSave
    def save(self, zip=False, quiet=False):

        t = time.time()

        bin, zip = OMFITmatrix.save(self, bin=True, zip=zip)

        if not quiet:
            if zip:
                z = 'zlibbed' if bin else 'gzipped'
            else:
                z = 'plain'
            _print(self, '%.2f s, %s, saved %s %s.' % (time.time() - t, sizeof_fmt(self.filename, ' '), z, 'NetCDF' if bin else 'ASCII'))

    def prep(self, project=False, force=False, save=None, prefix=True, quiet=False):

        if not force and np.all([w in self['data'].attrs for w in 'abc']):
            return
        if self.dynaLoad:
            self.load()

        # ignore all-nan warnings
        filterwarnings('ignore', r'All-NaN (slice|axis) encountered')

        if np.all(np.isnan(self['data'][:, 1, :].data)):
            project = True

        if np.any(np.isnan([self['rmax'], self['zmax']])) or np.any(np.isnan([self['rlim'], self['zlim'], self['slim']])):
            raise ValueError('missing gEQDSK data.')

        # shortcuts
        data = self['data'].data
        rlim = self['rlim']
        zlim = self['zlim']
        ss = self['slim']
        c = {v: k for k, v in enumerate(self['data'].coords['C'].data)}

        # total number of points
        tot = np.shape(data)[0]
        myformat = '{:' + str(len('{:,d}'.format(tot))) + ',d}'
        tstart = time.time()

        # hit points
        iih = np.zeros(tot).astype('int')
        rrh = np.nan * np.zeros(tot)
        zzh = np.nan * np.zeros(tot)
        phh = np.nan * np.zeros(tot)
        mah = np.nan * np.zeros(tot)

        # detect phi step
        dphi = np.abs(np.diff(data[:, 1:, c['phi']], axis=1)).squeeze()
        dphi[np.where(dphi[:] == 0.0)] = np.nan
        dpinit = np.nanmin(dphi)
        if not quiet:
            p = _print(self, (myformat + ' total points, step is {:.2f} deg.').format(tot, dpinit), prefix=prefix)

        # index lists
        ko = []
        try1 = []
        try2 = []

        # 1st TRY: USE SECOND-TO-LAST AND LAST POINT
        for t in range(tot):

            xx = data[t, 1:, c['rr']]
            yy = data[t, 1:, c['zz']]
            ph = data[t, 1:, c['phi']]
            ma = data[t, 1:, c['mlen']]

            try:

                # this is NOT the second-to-last point
                if np.isnan(dphi[t]) or dphi[t] > dpinit * 1.5:
                    raise ValueError

                # discard kicks greater than 50 cm
                if np.hypot(np.diff(xx), np.diff(yy)) > 0.5:
                    raise ValueError

                # compute intersection
                result, index = line_intersect(np.vstack((rlim, zlim)).T, np.vstack((xx, yy)).T, return_indices=True)
                rrh[t], zzh[t] = result[0]
                iih[t] = index[0][0]

            except (ValueError, IndexError):

                # discard second-to-last point
                ko.append(t)
                continue

            # compute normalized distance
            ddh = np.hypot(rrh[t] - xx[1], zzh[t] - yy[1])
            ddh /= np.hypot(xx[0] - xx[1], yy[0] - yy[1])

            # inteporlate phi, mag
            phh[t] = ph[1] + ddh * (ph[0] - ph[1])
            mah[t] = ma[1] + ddh * (ma[0] - ma[1])

            try1.append(t)

        # 2nd TRY: USE ONLY LAST POINT
        if project and np.size(ko):

            kos = np.array(ko).astype('int')
            ko = []

            # limiter-to-last point distance
            limd = np.nan * np.zeros((len(rlim), np.size(kos), 3))

            # for each divertor segment
            for l in range(len(rlim) - 1):

                if rlim[l] == rlim[l + 1] and zlim[l] == zlim[l + 1]:
                    continue

                # distance from segment center
                xc = np.mean([rlim[l], rlim[l + 1]])
                yc = np.mean([zlim[l], zlim[l + 1]])
                rc = np.hypot(xc - rlim[l], yc - zlim[l])
                (cpu,) = np.where(np.hypot(data[kos[:], 2, c['rr']] - xc, data[kos[:], 2, c['zz']] - yc) <= rc * 1.5)
                if not np.size(cpu):
                    continue

                # compute foot (xf,yf) and distance dd
                xf, yf, dd = foot_on_line(
                    xp=data[kos[cpu], 2, c['rr']], x1=rlim[l], x2=rlim[l + 1], yp=data[kos[cpu], 2, c['zz']], y1=zlim[l], y2=zlim[l + 1]
                )
                limd[l, cpu[:], 0] = xf[:]
                limd[l, cpu[:], 1] = yf[:]
                limd[l, cpu[:], 2] = dd[:]

            # for each point
            for k in range(len(kos)):
                (result,) = np.where(limd[:, k, 2] == np.nanmin(limd[:, k, 2]))
                try:
                    iih[kos[k]] = result[0]
                    try2.append(kos[k])
                except IndexError:
                    ko.append(kos[k])

            rrh[try2] = limd[iih[try2], list(range(len(try2))), 0]
            zzh[try2] = limd[iih[try2], list(range(len(try2))), 1]
            phh[try2] = data[try2, 2, c['phi']].copy()
            mah[try2] = data[try2, 2, c['mlen']].copy()

        # s: based on hit data
        ssh = ss[iih[:]] + np.hypot(rrh[:] - rlim[iih[:]], +zzh[:] - zlim[iih[:]])

        # store data
        self['data'].attrs['a'] = phh[:] * -1 % 360
        self['data'].attrs['b'] = ssh[:]
        self['data'].attrs['c'] = mah[:]

        # summary
        if np.size(try1):
            p = _print(self, (myformat + ' intersections.').format(np.size(try1)), tag='i', prefix=prefix, p=p)
        if np.size(try2):
            p = _print(self, (myformat + ' projections.').format(np.size(try2)), tag='w', prefix=prefix, p=p)
        if np.size(ko):
            p = _print(self, (myformat + ' errors.').format(np.size(ko)), tag='e', prefix=prefix, p=p)
        p = _print(self, 'computed hit data in {:.3f} s.'.format(time.time() - tstart), prefix=prefix, p=p)

        # save data
        if save is True or (save is None and self.OMFITproperties['bin'] is True):
            ro = self.readOnly
            self.readOnly = False
            self.save()
            self.readOnly = ro

    def plot(self, mlen=100, sol=False, cbar=True, inverse=False, prefix=True, quiet=False, **kw):
        r"""
        The OMFITtrip3Dhits.plot plots the footprint based on the data prepared by OMFITtrip3Dhits.prep.

        :param mlen: if None, select all the data points,
                     if MIN, select points with MIN <= mlen,
                     if (MIN,MAX),              MIN <= mlen <= MAX.

        :param cbar: if True,  the colorbar will be plotted,
                     if False,                      hidden.

        :param inverse: if False, y-axis is normal,
                        if True,            inverted.

        :param quiet: if False, output to console is normal,
                      if True,                       suppressed.

        :param \**kw: keyword dictionary to be passed to scatter().

        :return: None
        """

        filterwarnings('ignore', r'invalid value encountered in (less|greater)_equal')

        if not np.all([w in self['data'].attrs for w in 'abc']):
            self.prep(prefix=prefix, quiet=quiet)

        cc = self['data'].attrs['c']
        if sol and np.nanmin(cc) < mlen and mlen < np.nanmax(cc):

            i = isinteractive()
            if i:
                ioff()
            gs = GridSpec(2, 16)
            gs.update(wspace=1, hspace=0)
            ct = pyplot.subplot(gs[0, -1])
            cb = pyplot.subplot(gs[1, -1])
            ax = pyplot.subplot(gs[:, :-1])

            vmin = kw.pop('vmin', 0)
            vmax = kw.pop('vmax', None)
            cmap = kw.pop('cmap', None)
            if cmap is None or not np.iterable(cmap):
                cmapb = 'gist_earth_r'
                cmapt = 'plasma'
            else:
                cmapb = cmap[0]
                cmapt = cmap[1]

            self.plot(mlen=mlen, sol=False, cmap=cmapt, cbar=False, vmin=mlen, vmax=vmax, inverse=inverse, prefix=prefix, quiet=quiet, **kw)
            ct.set_title('len [m]')
            colorbar(cax=ct, format='%d', ticks=pyplot.MaxNLocator(nbins=5, prune='both'))

            self.plot(
                mlen=(vmin, mlen),
                sol=False,
                cmap=cmapb,
                cbar=False,
                vmin=vmin,
                vmax=mlen,
                inverse=inverse,
                prefix=prefix,
                quiet=quiet,
                **kw,
            )
            colorbar(cax=cb, format='%d', ticks=pyplot.MaxNLocator(nbins=5))

            if i:
                ion()
                draw()
                show()
            return ax, (cb, ct)

        # shortcuts
        p = False
        aa = self['data'].attrs['a']
        bb = self['data'].attrs['b']
        cc = self['data'].attrs['c']
        ss = self['slim']
        rlim = self['rlim']
        zlim = self['zlim']

        # total number of points
        tot = np.size(self['data'][:, 0])
        myformat = '{:' + str(len('{:,d}'.format(tot))) + ',d}'
        tstart = time.time()

        # magnetic length threshold
        if mlen is None:
            select = list(range(len(cc)))
        elif np.size(mlen) == 1:
            (select,) = np.where(cc >= mlen)
        elif np.iterable(mlen) and np.size(mlen) == 2:
            ave = np.mean(mlen)
            (dif,) = np.diff(mlen) / 2.0
            (select,) = np.where(abs(cc[:] - ave) <= dif)
        else:
            raise ValueError('unsupported magnetic length format: {:s} x {:d}.'.format(mlen.__class__.__name__, np.size(mlen)))
        if not np.size(select):
            if np.size(mlen) == 1:
                if not quiet:
                    p = _print(self, 'no points with length of {:.0f} m.'.format(mlen), tag='w', prefix=prefix, p=p)
                select = list(range(len(cc)))
            else:
                if not quiet:
                    p = _print(self, 'no points with length of [{:.0f}, {:.0f}] m.'.format(*mlen), tag='e', prefix=prefix, p=p)
                return None
        if not quiet:
            p = _print(
                self, 'magnetic length in [{:.0f}, {:.0f}] m.'.format(np.nanmin(cc[select]), np.nanmax(cc[select])), prefix=prefix, p=p
            )

        # limits
        pyplot.xlim(0, 360)
        pyplot.ylim(np.nanmin(bb[select]), np.nanmax(bb[select]))

        # default keywords
        kw.setdefault('cmap', 'plasma')
        kw.setdefault('s', 1)
        kw.setdefault('edgecolor', '')
        kw.setdefault('alpha', 0.75)
        kw.setdefault('rasterized', True)

        # plot
        scatter(aa[select], bb[select], c=cc[select], **kw)
        ax = pyplot.gca()

        # axes
        if inverse:
            pyplot.gca().invert_yaxis()
        pyplot.xlabel(r'toroidal angle, $\phi$ [deg]')
        pyplot.ylabel('hit parameter, $s$ [m]')
        ticklabel_format(useOffset=False)
        if cbar:
            cb = colorbar(format='%d', ticks=pyplot.MaxNLocator(prune='upper'))
            cb.ax.set_title('len [m]')

        # summary
        if not quiet:
            p = _print(
                self, 'plotted {:,d} points in {:.3f} s.'.format(np.sum(~np.isnan(bb[select])), time.time() - tstart), prefix=prefix, p=p
            )

        return ax, [cb.ax] if cbar else []


# Created by Trevisan at 01 May 2018  15:43
class OMFITtrip3Dstart(OMFITmatrix):

    """
    The OMFITtrip3Dstart class parses and handles the TRIP3D start points input file.
    A self-described xarray object is stored under self['data'].
    """

    @dynaLoad
    def load(self):

        t = time.time()

        OMFITmatrix.load(self, bin=False, zip=False, skiprows=1, comment='!')

        if self['data'] is not None:
            self.toxr()
            _print(self, '%.2f s, %s, load from file.' % (time.time() - t, sizeof_fmt(self.filename, ' ')))

    @dynaSave
    def save(self):

        t = time.time()

        na = np.shape(self['data'])[0]
        nb = np.max([1, np.size(np.where(self['data'][:, -1] == self['data'][0, -1]))])
        nc = na / nb
        with open(self.filename, 'w') as f:
            f.write(3 * '%d ' % (na, nb, nc) + '\n')

        OMFITmatrix.save(self, bin=False, zip=False, mode='a', na_rep='nan')

        _print(self, '%.2f s, %s, saved to file.' % (time.time() - t, sizeof_fmt(self.filename, ' ')))

    def toxr(self, data=None):

        if data is None:
            try:
                data = self['data'].values
            except Exception:
                data = self['data']

        self['data'] = xr.DataArray(data, name='startfile', dims=['P', 'C'], coords=[list(range(np.shape(data)[0])), ['rad', 'the', 'phi']])

    def plot(self, radvar=None, geqdsk=None, pol=True, tor=True, lim=True, surf=True, **kw):

        if radvar is None:
            radvar = eval(treeLocation(self)[-2])['trip3d.in']['input']['iRadVar']

        if geqdsk is None:
            try:
                geqdsk = eval(treeLocation(self)[-2])['gEQDSK']
            except Exception:
                pass
        try:
            rmax = geqdsk['RMAXIS']
            zmax = geqdsk['ZMAXIS']
        except Exception:
            rmax = np.nan
            zmax = np.nan
        try:
            rlim = geqdsk['RLIM']
            zlim = geqdsk['ZLIM']
        except Exception:
            rlim = np.nan
            zlim = np.nan

        suptitle(treeLocation(self)[-1])
        kw = {'rasterized': True}

        if pol:

            if tor:
                pyplot.subplot(122)
            pyplot.title('poloidal view')

            if radvar == 1:

                # rho
                rr = rmax + self['data'].loc[:, 'rad'] * np.cos(np.deg2rad(self['data'].loc[:, 'the']))
                zz = zmax + self['data'].loc[:, 'rad'] * np.sin(np.deg2rad(self['data'].loc[:, 'the']))

                if surf and geqdsk['NBBBS']:
                    geqdsk.plot(only2D=True)
                    pyplot.gca().autoscale(tight=True)
                    kw['color'] = pyplot.gca().lines[-2].get_color()
                if lim:
                    pyplot.plot(rlim, zlim, 'k-')
                    pyplot.gca().set_aspect('equal')

                pyplot.xlabel('R [m]')
                pyplot.ylabel('Z [m')

                pyplot.plot(rr, zz, '.', **kw)

            elif radvar == 3:

                # psi normal
                pyplot.plot(self['data'].loc[:, 'the'], self['data'].loc[:, 'rad'], '.', **kw)

                pyplot.xlabel('theta [deg]')
                pyplot.ylabel('psin')

        if tor:

            if pol:
                pyplot.subplot(121)
            pyplot.title('toroidal view')

            if radvar == 1:

                # rho
                rr = rmax + self['data'].loc[:, 'rad'] * np.cos(np.deg2rad(self['data'].loc[:, 'the']))
                pyplot.plot(
                    rr * np.cos(np.deg2rad(self['data'].loc[:, 'phi'])), rr * np.sin(np.deg2rad(self['data'].loc[:, 'phi'])), '.', **kw
                )

                if lim:
                    t = np.arange(361) * np.pi / 180.0
                    rmin = np.min(rlim)
                    rmax = np.max(rlim)
                    pyplot.plot(rmin * np.cos(t), rmin * np.sin(t), 'k-')
                    pyplot.plot(rmax * np.cos(t), rmax * np.sin(t), 'k-')

                pyplot.gca().set_aspect('equal')

            elif radvar == 3:

                # psi normal
                pyplot.plot(self['data'].loc[:, 'phi'], self['data'].loc[:, 'rad'], '.', **kw)

                pyplot.xlabel('phi [deg]')
                pyplot.ylabel('psin')


# Created by Trevisan at 01 May 2018  16:20
class OMFITtrip3Dprobeg(OMFITmatrix):

    """
    The OMFITtrip3Dprobeg class parses and handles the TRIP3D 'probe_gb.out' output file.
    A self-described xarray object is stored under self['data'].
    """

    @dynaLoad
    def load(self):

        t = time.time()

        OMFITmatrix.load(self, bin=False, zip=False, comment='%')

        if self['data'] is not None:
            self.toxr()
            _print(self, '%.2f s, %s, load from file.' % (time.time() - t, sizeof_fmt(self.filename, ' ')))

    @dynaSave
    def save(self):

        t = time.time()

        with open(self.filename, 'w') as f:
            f.write('%' + ' '.join(self['data'].coords['C'].values) + '\n')

        OMFITmatrix.save(self, bin=False, zip=False, mode='a', float_format='%.11e', na_rep='nan')

        _print(self, '%.2f s, %s, saved to file.' % (time.time() - t, sizeof_fmt(self.filename, ' ')))

    def toxr(self, data=None):

        if data is None:
            try:
                data = self['data'].values
            except Exception:
                data = self['data']

        self['data'] = xr.DataArray(
            data,
            name='probeg',
            dims=['P', 'C'],
            coords=[list(range(np.shape(data)[0])), ['phi', 'rr', 'zz', 'Bphi', 'Brr', 'Bzz', 'Bpol', 'Bmag']],
        )

    def plot(self, cols=['Bpol', 'Bmag'], phi=None, geqdsk=None, stats=True, **kw):

        if geqdsk is None:
            try:
                geqdsk = eval(treeLocation(self)[-3])['INPUTS']['gEQDSK']
            except Exception:
                pass
        try:
            rlim = geqdsk['RLIM']
            zlim = geqdsk['ZLIM']
        except Exception:
            rlim = np.nan
            zlim = np.nan

        if phi is None:
            phi = self['data'].loc[0, 'phi']
        (idx,) = np.where(self['data'].loc[:, 'phi'] == phi)
        if len(idx):
            printi('{:,d} points at phi = {:.1f} deg'.format(np.size(idx), phi.values))

        suptitle(treeLocation(self)[-1])

        kw.setdefault('cmap', 'plasma')
        kw.setdefault('s', 10)
        kw.setdefault('edgecolor', '')
        kw.setdefault('rasterized', True)

        i = 0
        for c in cols:

            if np.size(cols) > 1:
                i += 1
                pyplot.subplot(1, np.size(cols), i)

            if geqdsk:
                pyplot.plot(rlim, zlim, 'k-')

            scatter(self['data'].loc[idx, 'rr'], self['data'].loc[idx, 'zz'], c=self['data'].loc[idx, c], **kw)

            pyplot.gca().set_aspect('equal')
            pyplot.gca().autoscale(tight=True)
            pyplot.title(c)
            colorbar()

            if stats:
                printi(
                    c
                    + ' min = %g %s'
                    % (np.nanmin(self['data'].loc[idx, c]), '(NaNs excluded)' if np.isnan(self['data'].loc[idx, c]).any() else '')
                )
                printi(
                    ' ' * len(c)
                    + ' max = %g %s'
                    % (np.nanmax(self['data'].loc[idx, c]), '(NaNs excluded)' if np.isnan(self['data'].loc[idx, c]).any() else '')
                )


def _print(who, message, prefix=True, tag='i', p=False):

    # short versions of tag_print's tags
    tags = {
        'o': 'STDOUT',
        'e': 'STDERR',
        'd': 'DEBUG',
        'u': 'PROGRAM_OUT',
        'r': 'PROGRAM_ERR',
        'i': 'INFO',
        't': 'HIST',
        'h': 'HELP',
        'w': 'WARNING',
    }

    if prefix:

        # set the prefix
        if prefix == True:
            name = who.__class__.__name__
            method = extract_stack()[-2][2]
            if method == 'doplot':
                method = 'plot'
            if '__' in method:
                method = method[1:5]
            pref = '%s.%s' % (name, method)
        else:
            pref = str(prefix)
        message = '%s: %s' % (pref, message)

        # printed mask
        if p:
            msgs = message.split(': ')
            message = ' ' * len(': '.join(msgs[:-1])) + ': ' + msgs[-1]

    # print the message
    tag_print(message, tag=tags[tag])

    return True
